/*****************************************************************************
 *
 * FILE:	wdbm.h
 * DESCRIPTION:	wyvern database manager header
 * DATE:	Sat, Jun 30 2001
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/cover.html
 * COPYRIGHT:	(C) 2001 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *  Copyright (C) 2001-2002 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *  
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in
 *      the documentation and/or other materials provided with the
 *      distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY KOUICHI ABE (WALL) ``AS IS'' AND ANY
 *   EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL KOUICHI ABE (WALL) BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *   BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 *   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 *   OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 *   IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: wdbm.h,v 1.3 2002/01/08 16:29:53 kouichi Exp $
 *
 *****************************************************************************/

#ifndef	_wdbm_h
#define	_wdbm_h

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	USE_DB
#include <fcntl.h>
#include <db.h>
#elif	USE_NDBM
#include <fcntl.h>
#include <ndbm.h>
#elif	USE_GDBM
#include <gdbm.h>
#else
#include <fcntl.h>
#endif


/*****************************************************************************
 *
 *	definition of macros and structures
 *
 *****************************************************************************/
#if	USE_DB
typedef DB *	WDBM_FILE;
typedef struct {	/* The data and key structure */
  char *	dptr;	/* data */
  int		dsize;	/* data length */
} wdatum;
#elif	USE_DB2
typedef struct {
  DB *	dbp;
  DBC *	dbcp;
} _wdbm_file;
typedef _wdbm_file *	WDBM_FILE;
typedef struct {	/* The data and key structure */
  char *	dptr;	/* data */
  int		dsize;	/* data length */
} wdatum;
#elif	USE_NDBM
typedef DBM *		WDBM_FILE;
typedef	datum		wdatum;
#elif	USE_GDBM
typedef GDBM_FILE	WDBM_FILE;
typedef	datum		wdatum;
#else
#define	TEXT_DB	1
typedef struct {
  FILE *	fp;
  char *	name;	/* database name */
} _wdbm_file;
typedef _wdbm_file *	WDBM_FILE;
typedef struct {	/* The data and key structure */
  char *	dptr;	/* data */
  size_t	dsize;	/* data length */
} wdatum;
#endif

/* database open mode */
#define	WDBM_READER	0	/* A reader. */
#define	WDBM_WRITER	1	/* A writer. */
#define	WDBM_WRCREAT	2	/* A writer. Create the db if needed. */
#define	WDBM_NEWDB	3	/* A writer. Always create a new db. */

/* database store flag */
#define	WDBM_INSERT	0	/* Never replace old data with new. */
#define	WDBM_REPLACE	1	/* Always replace old data with new. */

/**/
extern WDBM_FILE	wdbm_open(const char *, int, int);
extern void		wdbm_close(WDBM_FILE);
extern int		wdbm_store(WDBM_FILE, wdatum, wdatum, int);
extern wdatum		wdbm_fetch(WDBM_FILE, wdatum);
extern int		wdbm_delete(WDBM_FILE, wdatum);
extern wdatum		wdbm_firstkey(WDBM_FILE);
extern wdatum		wdbm_nextkey(WDBM_FILE, wdatum);
extern int		wdbm_exists(WDBM_FILE, wdatum);

#endif	/* _wdbm_h */
