/*****************************************************************************
 *
 * FILE:	alias.c
 * DESCRIPTION:	register Redirect/Alias/ScriptAlias
 * DATE:	Sun, Sep 26 1999
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/cover.html
 * COPYRIGHT:	(C) 1999-2003 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *   Copyright (C) 1999-2003 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *   All rights reserved.
 *  
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions
 *   are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in
 *      the documentation and/or other materials provided with the
 *      distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY KOUICHI ABE (WALL) ``AS IS'' AND ANY
 *   EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL KOUICHI ABE (WALL) BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *   BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 *   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 *   OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 *   IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: alias.c,v 1.50.2.13 2003/03/18 16:06:05 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#if	HAVE_STDLIB_H
#include <stdlib.h>
#endif	/* HAVE_STDLIB_H */
#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#include "wyvern.h"
#include "functions.h"


/*****************************************************************************
 *
 *	declaration of local variables
 *
 *****************************************************************************/


/*****************************************************************************
 *
 *	declaration of local functions
 *
 *****************************************************************************/
static String	lookup(Alias *, const String, alias_t);


/*****************************************************************************
 *
 *	definition of functions
 *
 *****************************************************************************/

/*
 * Function Name:	aliasEntry
 * Description:		register a Redirect, Alias, or ScriptAlias into the
 *			aliasTable.
 * Arguments:		aliasTable, fakeName, realName, alias_type
 * Return Value:	none
 */
void
aliasEntry(aliasTable, fakeName, realName, type)
	Alias **	aliasTable;
	const String	fakeName;
	const String	realName;
	alias_t		type;
{
  Alias *	new;

  if (lookup(*aliasTable, fakeName, type) != NULL) {
    return;	/* already registered */
  }

  new		= (Alias *)wvMalloc(sizeof(Alias));
  new->type	= type;
  new->fakeName	= wvStrdup(fakeName);
  new->fakeLen	= strlen(fakeName);
  new->realName	= wvStrdup(realName);
  new->realLen	= strlen(realName);
  new->next	= *aliasTable;
  *aliasTable	= new;
}

/*
 * Function Name:	lookup
 * Description:		lookup the Redirect, Alias, or ScriptAlias from the
 *			aliasTable.
 * Arguments:		aliasTable, fakeName, alias_type
 * Return Value:	realName
 */
static String
lookup(aliasTable, fakeName, type)
	Alias *		aliasTable;
	const String	fakeName;
	alias_t		type;
{
  register Alias *	ap;

  for (ap = aliasTable; ap != NULL; ap = ap->next) {
    if (wvStrncmp(fakeName, ap->fakeName, ap->fakeLen) == 0) {
      return ap->realName;
    }
  }
  return NULL;
}


/*
 * Function Name:	translateAlias
 * Description:		translate an aliases of request URI
 * Arguments:		message
 * Return Value:	successful: 0, no translated: 1, otherwise: -1
 */
int
translateAlias(msg)
	register Message *	msg;
{
  register String	urip;
  register Alias *	ap;

#if	DEBUG
  fputs("<translateAlias>\n", stderr);
#endif	/* DEBUG */
  /* find ScriptAlias, Alias, or Redirect */
  for (ap = msg->vhost->aliasTable; ap != NULL; ap = ap->next) {
    int	n;

    if (wvStrncmp(msg->request.uri, ap->fakeName, ap->fakeLen) == 0 ||
	wvStrncmp(msg->request.uri, ap->fakeName, ap->fakeLen - 1) == 0) {
#if	DEBUG
      fprintf(stderr, "[translateAlias] checking %s\n", ap->fakeName);
#endif	/* DEBUG */
      switch (ap->type) {
	/* Redirection */
	case A_REDIRECT:
	  if (responseHeaderEntry(&msg->response.headers,
				  "Location", ap->realName) == NULL) {
	    goto error;
	  }
	  msg->response.status = STATUS_MOVED_PERMANENTLY;
#if	DEBUG
	  fprintf(stderr, "[translateAlias] %s\n", ap->realName);
#endif	/* DEBUG */
	  return RET_OK;
	/* Alias */
	case A_ALIAS:
	  n = asprintf(&msg->file.path, "%s%s",
			ap->realName, msg->request.uri + ap->fakeLen);
	  if (n == -1) {
	    goto error;
	  }
#if	DEBUG
	  fprintf(stderr, "[translateAlias] %s\n", msg->file.path);
#endif	/* DEBUG */
	  return RET_OK;
	/* Script Alias */
	case A_SCRIPT:
	  urip = (String)strchr(msg->request.uri + ap->fakeLen, '/');
	  if (urip != NULL) {
	    msg->request.path_info = wvStrdup(urip);	*urip = EOL;
	    if (msg->request.path_info == NULL) {
	      goto error;
	    }
	  }
	  n = asprintf(&msg->file.path, "%s%s",
			ap->realName, msg->request.uri + ap->fakeLen);
	  if (n == -1) {
	    goto error;
	  }
	  /* directory listing for cgi-bin directory */
	  if (msg->file.path[n - 1] == '/') {
	    msg->response.status = STATUS_BAD_REQUEST;
	    return RET_NG;
	  }
	  msg->file.mime = CGI_MAGIC_TYPE;
#if	DEBUG
	  fprintf(stderr, "[translateScript] %s\n", msg->file.path);
#endif	/* DEBUG */
	  return RET_OK;
	/* ignore */
	default:
	  break;
      }
    }
  }

  return RET_CONT;	/* no translated URI */

error:
  return wvNoMemoryError(msg);
}

