%{
/*****************************************************************************
 *
 * FILE:	conf_grammar.y
 * DESCRIPTION:	configuration parser
 * DATE:	Thu, Sep  9 1999
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/cover.html
 * COPYRIGHT:	(C) 1999-2002 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *  Copyright (C) 1999-2002 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *  
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in
 *      the documentation and/or other materials provided with the
 *      distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY KOUICHI ABE (WALL) ``AS IS'' AND ANY
 *   EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL KOUICHI ABE (WALL) BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *   BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 *   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 *   OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 *   IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: conf_grammar.y,v 1.22.2.7 2002/09/25 01:23:50 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#include <stdlib.h>
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#include <string.h>
#include <sys/types.h>
#include <pwd.h>
#include <grp.h>
#include "wyvern.h"
#include "functions.h"

struct passwd *	pbuf;
struct group *	gbuf;

static char	bindAddr[16];
static char	mimeType[BUFSIZ];
static char	fakeName[BUFSIZ];

static char	flag = 0;
%}

%union {
  int		ival;	/* INTEGER	*/
  char *	sval;	/* STRING	*/
};

/* wyvern.conf tokens	*/
%token W_LISTEN W_MAXTHREAD W_MAXREQUEST
%token W_USER W_GROUP W_SERVERROOT W_SECURELEVEL W_SECURITYLOG W_PIDFILE
%token W_TMPDIR W_CACHEDIR
%token W_SERVERNAME W_SERVERADMIN W_ERRORLOG W_ACCESSLOG W_TIMEOUT

%token W_DOCUMENTROOT W_USERDIR W_DIRECTORYINDEXES W_AUTOINDEXING
%token W_FOLLOWSYMLINKS W_DEFAULTTYPE W_ERROR_DOCUMENT
%token W_REDIRECT W_ALIAS W_SCRIPTALIAS W_ADDTYPE W_MODULE

%token W_BEGIN_VIRTUALHOST W_END_VIRTUALHOST

%token <ival> INTEGER
%token <sval> STRING
%token <sval> IPADDR
%token <sval> MIMETYPE

%%

ConfigFiles:		WyvernConfigStmts MimeTypeStmts
	;

WyvernConfigStmts:	WyvernConfigStmts WyvernConfigStmt
	|		/* empty */
	;

WyvernConfigStmt:	ListenStmt
	|		MaxThreadStmt
	|		MaxRequestStmt
	|		UserStmt
	|		GroupStmt
	|		ServerRootStmt
	|		SecureLevelStmt
	|		SecurityLogStmt
	|		PidFileStmt
	|		TmpDirStmt
	|		CacheDirStmt
	|		ServerNameStmt
	|		ServerAdminStmt
	|		ErrorLogStmt
	|		AccessLogStmt
	|		TimeoutStmt
	|		DocumentRootStmt
	|		UserDirStmt
	|		DirectoryIndexesStmt
	|		AutoIndexingStmt
	|		FollowSymLinksStmt
	|		DefaultTypeStmt
	|		ErrorDocumentStmt
	|		RedirectStmt
	|		AliasStmt
	|		ScriptAliasStmt
	|		AddTypeStmt
	|		ModuleStmt
	|		BeginVirtualHostStmt
	|		EndVirtualHostStmt
	;

ListenStmt:		W_LISTEN BindAddressStmt
	;

BindAddressStmt:	PortStmt
	|		PortStmt ':' SecurePortStmt
	|		AddressStmt ':' PortStmt
	|		AddressStmt ':' PortStmt ':' SecurePortStmt
	;

AddressStmt:		IPADDR
	{
	  memset(bindAddr, 0, sizeof(bindAddr));
	  wvStrncpy(bindAddr, $1, sizeof(bindAddr));
	  flag = 0x4;
	}
	;

PortStmt:		INTEGER
	{
	  if (flag & 0x4) {	/* bind any address */
	    socketEntry(bindAddr, $1, SOC_UNSECURE);
	    flag = 0x2;
	  }
	  else {
	    socketEntry(NULL, $1, SOC_UNSECURE);
	  }
	}
	;

SecurePortStmt:		INTEGER
	{
	  if (flag & 0x2) {	/* bind any address */
	    socketEntry(bindAddr, $1, SOC_SECURE);
	    flag = 0x0;	/* reset */
	  }
	  else {
	    socketEntry(NULL, $1, SOC_SECURE);
	  }
	}
	;

MaxThreadStmt:		W_MAXTHREAD INTEGER
	{
	  httpd->maxThread = ($2 < 0 ? DEF_MIN_THREAD : $2);
	}
	;

MaxRequestStmt:		W_MAXREQUEST INTEGER
	{
	  httpd->maxRequest = ($2 < 0 ? DEF_MIN_REQUEST : $2);
	}
	;

UserStmt:		W_USER STRING
	{
	  pbuf = getpwnam($2);
	  if (pbuf == NULL) {
	    fprintf(stderr, "No such user: %s\n", $2);
	    exit(1);
	  }
	  httpd->user = pbuf->pw_uid;
	}
	|		W_USER INTEGER
	{
	  pbuf = getpwuid((uid_t)$2);
	  if (pbuf == NULL) {
	    fprintf(stderr, "No such user ID: %d\n", $2);
	    exit(1);
	  }
	  httpd->user = pbuf->pw_uid;
	}
	;

GroupStmt:		W_GROUP STRING
	{
	  gbuf = getgrnam($2);
	  if (gbuf == NULL) {
	    fprintf(stderr, "No such group: %s\n", $2);
	    exit(1);
	  }
	  httpd->group = gbuf->gr_gid;
	}
	|		W_GROUP INTEGER
	{
	  gbuf = getgrgid((gid_t)$2);
	  if (gbuf == NULL) {
	    fprintf(stderr, "No such group ID: %d\n", $2);
	    exit(1);
	  }
	  httpd->group = gbuf->gr_gid;
	}
	;

ServerRootStmt:		W_SERVERROOT STRING
	{
	  httpd->serverRoot = wvStrdup($2);
	  chdir(httpd->serverRoot);	/* for mime.types */
	}
	;

SecureLevelStmt:	W_SECURELEVEL INTEGER
	{
	  httpd->secureLevel = $2;
	}
	;

SecurityLogStmt:	W_SECURITYLOG STRING
	{
	  httpd->securityLog.path = wvStrdup($2);
	}
	;

PidFileStmt:		W_PIDFILE STRING
	{
	  httpd->pidFile = wvStrdup($2);
	}
	;

TmpDirStmt:		W_TMPDIR STRING
	{
	  httpd->tmpDir = wvStrdup($2);
	}
	;

CacheDirStmt:		W_CACHEDIR STRING
	{
	  httpd->cacheDir = wvStrdup($2);
	}
	;

ServerNameStmt:		W_SERVERNAME STRING
	{
	  httpd->vhost->serverName = wvStrdup($2);
	}
	;

ServerAdminStmt:	W_SERVERADMIN STRING
	{
	  httpd->vhost->serverAdmin = wvStrdup($2);
	}
	;

ErrorLogStmt:		W_ERRORLOG STRING
	{
	  httpd->vhost->errorLog.path = wvStrdup($2);
	}
	;

AccessLogStmt:		W_ACCESSLOG STRING
	{
	  httpd->vhost->accessLog.path = wvStrdup($2);
	}
	;

TimeoutStmt:		W_TIMEOUT INTEGER
	{
	  if ($2 < 0) {
	    fputs("Timeout must be positive value\n", stderr);
	    exit(1);
	  }
	  httpd->vhost->timeout = $2;
	}
	;

DocumentRootStmt:	W_DOCUMENTROOT STRING
	{
	  httpd->vhost->docRoot = wvStrdup($2);
	}
	;

UserDirStmt:		W_USERDIR STRING
	{
	  httpd->vhost->userDir = wvStrdup($2);
	}
	;

DirectoryIndexesStmt:	W_DIRECTORYINDEXES STRING
	{
	  httpd->vhost->dirIndexes = indexesEntry($2);
	}
	;

AutoIndexingStmt:	W_AUTOINDEXING STRING
	{
	  if (wvStrcasecmp($2, "on") == 0) {
	    httpd->vhost->autoIndexing = TRUE;
	  }
	  else {
	    httpd->vhost->autoIndexing = FALSE;
	  }
	}
	;

FollowSymLinksStmt:	W_FOLLOWSYMLINKS STRING
	{
	  if (wvStrcasecmp($2, "on") == 0) {
	    httpd->vhost->symlinks = TRUE;
	  }
	  else {
	    httpd->vhost->symlinks = FALSE;
	  }
	}
	;

DefaultTypeStmt:	W_DEFAULTTYPE STRING
	{
	  httpd->vhost->defaultType = wvStrdup($2);
	}
	;

ErrorDocumentStmt:	W_ERROR_DOCUMENT INTEGER STRING
	{
	  errorDocEntry(&httpd->vhost->errorDocTable, $2, $3);
	}
	;

RedirectStmt:		W_REDIRECT STRING
	{
	  memset(fakeName, 0, sizeof(fakeName));
	  wvStrncpy(fakeName, $2, sizeof(fakeName));
	}
	STRING
	{
	  aliasEntry(&httpd->vhost->aliasTable, fakeName, $4, A_REDIRECT);
	}
	;

AliasStmt:		W_ALIAS STRING
	{
	  memset(fakeName, 0, sizeof(fakeName));
	  wvStrncpy(fakeName, $2, sizeof(fakeName));
	}
	STRING
	{
	  aliasEntry(&httpd->vhost->aliasTable, fakeName, $4, A_ALIAS);
	}
	;

ScriptAliasStmt:	W_SCRIPTALIAS STRING
	{
	  memset(fakeName, 0, sizeof(fakeName));
	  wvStrncpy(fakeName, $2, sizeof(fakeName));
	}
	STRING
	{
	  aliasEntry(&httpd->vhost->aliasTable, fakeName, $4, A_SCRIPT);
	}
	;

AddTypeStmt:		W_ADDTYPE MimeTypeStmt
	;

ModuleStmt:		W_MODULE STRING
	{
	  moduleLoad($2);
	}
	;

BeginVirtualHostStmt:	W_BEGIN_VIRTUALHOST
	{
	  httpd->vhost = vhostNew(yylval.sval);
	}
	;

EndVirtualHostStmt:	W_END_VIRTUALHOST
	{
	  if (httpd->vhost->serverAdmin == NULL) {
	    fprintf(stderr, "No ServerAdmin in VirtualHost: %s\n",
		    httpd->vhost->serverName);
	    exit(1);
	  }
	  if (httpd->vhost->docRoot == NULL) {
	    fprintf(stderr, "No DocumentRoot in VirtualHost: %s\n",
		    httpd->vhost->serverName);
	    exit(1);
	  }
	  httpd->vhost = httpd->host;
	}
	;

/*
 *	mime.types
 */

MimeTypeStmts:		MimeTypeStmts MimeTypeStmt
	|		/* empty */
	;

MimeTypeStmt:		MIMETYPE
	{
	  register char *	p = $1;
	  if (*p == '"') {
	    p++;
	    p[strlen(p) - 1] = '\0';
	  }
	  memset(mimeType, 0, sizeof(mimeType));
	  wvStrncpy(mimeType, p, sizeof(mimeType));
	}
	SuffixList
	;

SuffixList:		SuffixList Suffix
	|		/* empty */
	;

Suffix:			STRING
	{
	  mimeTypeEntry($1, mimeType);
	}
	;

%%
