/*****************************************************************************
 *
 * FILE:	config.c
 * DESCRIPTION:	read configuration file
 * DATE:	Wed, Sep  8 1999
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/cover.html
 * COPYRIGHT:	(C) 1999-2003 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *   Copyright (C) 1999-2003 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *   All rights reserved.
 *  
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions
 *   are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in
 *      the documentation and/or other materials provided with the
 *      distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY KOUICHI ABE (WALL) ``AS IS'' AND ANY
 *   EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL KOUICHI ABE (WALL) BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *   BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 *   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 *   OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 *   IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: config.c,v 1.27.2.6 2003/03/18 16:06:06 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include "wyvern.h"
#include "functions.h"


/*****************************************************************************
 *
 *	definition of local functions
 *
 *****************************************************************************/


/*****************************************************************************
 *
 *	definition of functions
 *
 *****************************************************************************/

/*
 * Function Name:	readConfigFile
 * Description:		read the configuration file via yyparse.
 * Arguments:		none
 * Return Value:	none
 */
void
readConfigFile(void)
{
  extern FILE *		yyin;
  extern int		yyparse(void);

  /* open config file */
  yyin = wvSafeFOpen(httpd->configFile, "r");
  if (yyin == NULL) {
    switch (errno) {
      case ENOENT:
	die("Not exist: %s\n", httpd->configFile);
      case EACCES:
	die("failed to access %s\n", httpd->configFile);
      case EMLINK:
	die("%s: untrusted symbolic link\n", httpd->configFile);
      default:
	die("failed to open %s\n", httpd->configFile);
    }
  }

  /*
   * yyparse() returns
   *	 0: if parsing is successful
   *	 1: if parsing failed
   */
  if (yyparse()) {
    die("failed to parse %s\n", httpd->configFile);
  }
  fclose(yyin);

  /* verify configuration */
  if (httpd->host->serverName == NULL) {
    httpd->host->serverName = getServerName();
  }
  if (httpd->host->serverAdmin == NULL) {
    die("No ServerAdmin in wyvern.conf\n");
  }

  /* useful trick */
  httpd->host->next = httpd->vhostTable;
}

