/*****************************************************************************
 *
 * FILE:	daemon.c
 * DESCRIPTION:	make a daemon
 * DATE:	Thu, Sep 14 1999
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/cover.html
 * COPYRIGHT:	(C) 1999-2003 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *   Copyright (C) 1999-2003 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *   All rights reserved.
 *  
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions
 *   are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in
 *      the documentation and/or other materials provided with the
 *      distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY KOUICHI ABE (WALL) ``AS IS'' AND ANY
 *   EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL KOUICHI ABE (WALL) BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *   BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 *   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 *   OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 *   IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: daemon.c,v 1.17.2.11 2003/03/18 16:06:06 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#if	HAVE_FCNTL_H
#include <fcntl.h>
#endif	/* HAVE_FCNTL_H */
#if	HAVE_SYS_TIME_H
#include <sys/time.h>
#endif	/* HAVE_SYS_TIME_H */
#include <sys/resource.h>
#include "wyvern.h"
#include "functions.h"


/*****************************************************************************
 *
 *	definition of macros and structures
 *
 *****************************************************************************/


/*****************************************************************************
 *
 *	declaration of local functions
 *
 *****************************************************************************/
static void	systemResourceLimiter(void);


/*****************************************************************************
 *
 *	definition of functions
 *
 *****************************************************************************/

/*
 * Function Name:	detach
 * Description:		detach the controlling terminal and create a new
 *			session
 * Arguments:		none
 * Return Value:	none
 */
void
detach(void)
{
  int	fdmax = sysconf(_SC_OPEN_MAX);
  int	fd;

#if	DEBUG
  /* close all descriptor */
  close(1);	/* stdout */
  close(0);	/* stdin */
  for (fd = 3; fd < fdmax; fd++) {
    close(fd);
  }

  /* reopen stdin/stdout */
  fd = open("/dev/null", O_RDWR);
  dup2(1, fd);	/* stdout */
  dup2(0, fd);	/* stdin */

  /* limit system resources */
  systemResourceLimiter();

  /* change working directory */
  (void)chdir(httpd->serverRoot);
#else
  pid_t	pid;

  /* limit system resources */
  systemResourceLimiter();

  /* detach the current process from shell */
  pid = fork();
  switch (pid) {
    case -1:	/* error */
      exit(1);
    case  0:	/*  child process */
      break;	/* continue */
    default:	/* parent process */
      _exit(0);
  }

  /* create new session */
  (void)setsid();

  /* detach the process from the 1st child process */
  switch (fork()) {
    case -1:	/* error */
      exit(1);
    case  0:	/*  child process */
      break;	/* continue */
    default:	/* parent process */
      _exit(0);
  }

  /* change working directory */
  (void)chdir(httpd->serverRoot);

  /* clear the file mode mask */
  (void)umask(022);

#if	0
  /* close all descriptor */
  for (fd = 0; fd < fdmax; fd++) {
    close(fd);
  }

  /* reopen stdin/stdout */
  fd = open("/dev/null", O_RDWR);
  dup(fd);
  dup(fd);
#else
  (void)freopen("/dev/null", "w", stdout);	/* ignore error */
  (void)freopen("/dev/null", "r", stdin);	/* ignore error */
#endif
#endif	/* DEBUG */
}

/*****************************************************************************/

/*
 * Function Name:	storePID
 * Description:		store the process id into the pid_file
 * Arguments:		none
 * Return Value:	none
 */
void
storePID(void)
{
  register FILE *	fout;

  /* store wyvern pid */
  fout = wvSafeFOpen(httpd->pidFile, "w");
  if (fout == NULL) {
    die("failed to open %s\n", httpd->pidFile);
  }
  httpd->pid = getpid();	/* get current pid */
  fprintf(fout, "%ld\n", (long)httpd->pid);
  if (getuid() == UID_ROOT) {
    /* change owner/group of process ID file */
    (void)fchown(fileno(fout), httpd->user, httpd->group);
  }
  fclose(fout);
}

/*
 * Function Name:	systemResourceLimiter
 * Description:		restrict the system resources of wyvern process
 * Arguments:		none
 * Return Value:	none
 */
static void
systemResourceLimiter(void)
{
  struct rlimit	rl;

  if (getrlimit(RLIMIT_CORE, &rl) == 0) {
    rl.rlim_cur = rl.rlim_max = 0;	/* no core dump */
    setrlimit(RLIMIT_CORE, &rl);
  }
  if (getrlimit(RLIMIT_NOFILE, &rl) == 0) {
    rl.rlim_cur = rl.rlim_max;
    setrlimit(RLIMIT_NOFILE, &rl);
  }
#if	0
  if (getrlimit(RLIMIT_FSIZE, &rl) == 0) {
    rl.rlim_cur = MAX_FILE_SIZE;
    setrlimit(RLIMIT_FSIZE, &rl);
  }
#endif
}

