/*****************************************************************************
 *
 * FILE:	log.c
 * DESCRIPTION:	record server logging
 * DATE:	Sun, Sep 26 1999
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/cover.html
 * COPYRIGHT:	(C) 1999-2003 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *   Copyright (C) 1999-2003 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *   All rights reserved.
 *  
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions
 *   are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in
 *      the documentation and/or other materials provided with the
 *      distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY KOUICHI ABE (WALL) ``AS IS'' AND ANY
 *   EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL KOUICHI ABE (WALL) BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *   BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 *   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 *   OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 *   IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: log.c,v 1.35.2.25 2003/08/18 12:57:45 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_STDLIB_H
#include <stdlib.h>
#endif	/* HAVE_STDLIB_H */
#if	__STDC__
#include <stdarg.h>
#else
#include <varargs.h>
#endif	/* __STDC__ */
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#include <time.h>
#include "wyvern.h"
#include "functions.h"


/*****************************************************************************
 *
 *	definition of macros and structures
 *
 *****************************************************************************/
#ifndef	HAVE_SETLINEBUF
#define	setlinebuf(x)	setvbuf((x), (char *)NULL, _IOLBF, 0)
#endif	/* HAVE_SETLINEBUF */


/*****************************************************************************
 *
 *	declaration of local variables
 *
 *****************************************************************************/


/*****************************************************************************
 *
 *	declaration of local functions
 *
 *****************************************************************************/
static void	openLog(register VirtualHost *);
static void	subsyslog(log_t, const char *, va_list);


/*****************************************************************************
 *
 *	definition of functions
 *
 *****************************************************************************/

/*
 * Function Name:	openLogFiles
 * Description:		open log files
 * Arguments:		none
 * Return Value:	none
 */
void
openLogFiles(void)
{
  register VirtualHost *	vp;

  /* default host */
  httpd->vhost = NULL;
  openLog(httpd->host);
  httpd->vhost = httpd->host;

  /* virtual hosts */
  for (vp = httpd->vhostTable; vp != NULL; vp = vp->next) {
    if (wvStrcmp(vp->serverName, DEF_VHOSTNAME) == 0) {
      continue;
    }
    openLog(vp);
    /* open error-log file */
    if (vp->errorLog.path == NULL) {
      vp->errorLog.fp = httpd->host->errorLog.fp;
    }
    /* open access-log file */
    if (vp->accessLog.path == NULL) {
      vp->accessLog.fp = httpd->host->accessLog.fp;
    }
  }

  /* security log */
  if (httpd->securityLog.path != NULL) {
    if (httpd->securityLog.path[0] == '|') {
      httpd->securityLog.fp = popen(&httpd->securityLog.path[1], "w");
    }
    else {
      httpd->securityLog.fp = wvSafeFOpen(httpd->securityLog.path, "a");
    }
    if (httpd->securityLog.fp == NULL) {
      die("failed to open securityLog: %s\n", httpd->securityLog.path);
    }
    (void)setlinebuf(httpd->securityLog.fp);	/* ignore error */
  }
}

/*
 * Function Name:	openLog
 * Description:		open log files
 * Arguments:		virtual host list
 * Return Value:	none
 */
static void
openLog(vp)
	register VirtualHost *	vp;
{
  /* errorLog */
  if (vp->errorLog.path != NULL) {
    if (vp->errorLog.path[0] == '|') {
      vp->errorLog.fp = popen(&vp->errorLog.path[1], "w");
    }
    else {
      vp->errorLog.fp = wvSafeFOpen(vp->errorLog.path, "a");
    }
    if (vp->errorLog.fp == NULL) {
      die("failed to open errorLog: %s\n", vp->errorLog.path);
    }
    (void)setlinebuf(vp->errorLog.fp);	/* ignore error */
  }

  /* accessLog */
  if (vp->accessLog.path != NULL) {
    if (vp->accessLog.path[0] == '|') {
      vp->accessLog.fp = popen(&vp->accessLog.path[1], "w");
    }
    else {
      vp->accessLog.fp = wvSafeFOpen(vp->accessLog.path, "a");
    }
    if (vp->accessLog.fp == NULL) {
      die("failed to open accessLog: %s\n", vp->accessLog.path);
    }
    (void)setlinebuf(vp->accessLog.fp);	/* ignore error */
  }
}

/*
 * Function Name:	closeLogFiles
 * Description:		close log files
 * Arguments:		none
 * Return Value:	none
 */
void
closeLogFiles(void)
{
  register VirtualHost *	vp;

  /* access/error log */
  for (vp = httpd->host; vp != NULL; vp = vp->next) {
    if (vp->errorLog.fp == NULL || vp->accessLog.fp == NULL) {
      continue;
    }
    closeLog(vp);
  }

  /* security log */
  if (httpd->securityLog.path != NULL && httpd->securityLog.fp != NULL) {
    if (httpd->securityLog.path[0] == '|') {
      pclose(httpd->securityLog.fp);
    }
    else {
      fclose(httpd->securityLog.fp);
    }
    httpd->securityLog.fp = NULL;
  }
}

/*
 * Function Name:	closeLog
 * Description:		close log files
 * Arguments:		virtual host list
 * Return Value:	none
 */
void
closeLog(vp)
	register VirtualHost *	vp;
{
  /* close error-log file */
  if (vp->errorLog.path != NULL) {
    if (vp->errorLog.path[0] == '|') {
      pclose(vp->errorLog.fp);
    }
    else {
      fclose(vp->errorLog.fp);
    }
    vp->errorLog.fp = NULL;
  }

  /* close access-log file */
  if (vp->accessLog.path != NULL) {
    if (vp->accessLog.path[0] == '|') {
      pclose(vp->accessLog.fp);
    }
    else {
      fclose(vp->accessLog.fp);
    }
    vp->accessLog.fp = NULL;
  }
}

/*****************************************************************************/

/*
 * Function Name:	defaultLogWriter
 * Description:		write the log message in log file
 * Arguments:		message
 * Return Value:	successful: 0, otherwise: -1
 */
int
defaultLogWriter(msg)
	register Message *	msg;
{
  String	client_name;
  
  client_name = msg->conn->client.name == NULL ? msg->conn->client.ip
					       : msg->conn->client.name;
  /* Client Error: 4xx */
  /* Server Error: 5xx */
  if (msg->response.status > STATUS_CODE_4XX) {
    if (msg->request.x_forwarded_for != NULL &&
	wvStrncasecmp(msg->request.x_forwarded_for, "unknown", 7) == 0) {
      wvSyslog(LOG_WARNING,
		"(%s) \"%s %s %s\" request forwarded from unknown host.\n",
		client_name,
		methodName[msg->request.method],
		msg->request.uri,
		httpVer[msg->request.http]);
    }
    else {
      log_t	log_type = LOG_ERR;

      if (msg->response.status == STATUS_FORBIDDEN ||
	  msg->response.status == STATUS_UNAUTHORIZED) {
	log_type |= LOG_SECURITY;
      }
      wvSyslog(log_type, "%s \"%s %s %s\" %s (%d)\n",
		client_name,
		methodName[msg->request.method],
		msg->request.uri,
		httpVer[msg->request.http],
		responseStatusTable[msg->response.status].phrase,
		responseStatusTable[msg->response.status].code);
    }
  }
  /* Informational: 1xx */
  /* Successful:    2xx */
  /* Redirection:   3xx */
  else {
    if (msg->request.x_forwarded_for != NULL) {
      fprintf(msg->vhost->accessLog.fp, "%s(%s) ",
	      client_name, msg->request.x_forwarded_for);
    }
    else {
      fprintf(msg->vhost->accessLog.fp, "%s ", client_name);
    }
    fprintf(msg->vhost->accessLog.fp, "%s %s [%s] \"%s %s %s\" %d ",
	    msg->user.identName == NULL ? "-" : msg->user.identName,
	    msg->user.authName == NULL ? "-" : msg->user.authName,
	    getTimeOfRFC822(msg->accesstime, TIMEZONE_LOCAL),
	    methodName[msg->request.method],
	    msg->request.uri,
	    httpVer[msg->request.http],
	    responseStatusTable[msg->response.status].code);
    if (msg->response.status > STATUS_CODE_3XX) {
	fputs("-", msg->vhost->accessLog.fp);
    }
    else {
      switch (msg->request.method) {
	case M_HEAD:
	  fputs("0", msg->vhost->accessLog.fp);
	  break;
	case M_GET:
	  fprintf(msg->vhost->accessLog.fp, "%ld",
		  (long)msg->response.content_length);
	  break;
	case M_POST:
	  fprintf(msg->vhost->accessLog.fp, "%d",
		  msg->request.content_length);
	  break;
	default:
	fputs("-", msg->vhost->accessLog.fp);
	  break;
      }
    }
#if	USE_SSL
    if (msg->conn->sock.type == SOC_SECURE) {
      fprintf(msg->vhost->accessLog.fp, " %s %s",
	      SSL_get_version(msg->conn->sock.ssl),
	      SSL_get_cipher(msg->conn->sock.ssl));
    }
#endif	/* USE_SSL */
    fputs("\n", msg->vhost->accessLog.fp);
  }

  return RET_OK;
}

/*****************************************************************************/

/*
 * Function Name:	die
 * Description:		print an error message and terminate process
 * Arguments:		type of error
 * Return Value:	none
 */
void
#if	__STDC__
die(const char *fmt, ...)
#else
die(fmt, va_alist)
	char *	fmt;
	va_dcl
#endif	/* __STDC__ */
{
  va_list	ap;
#if	__STDC__
  va_start(ap, fmt);
#else
  va_start(ap);
#endif	/* __STDC__ */
  subsyslog(LOG_FATAL, fmt, ap);
  va_end(ap);

  if (httpd->pid > 0) {
    (void)unlink(httpd->pidFile);	/* ignore error */
  }
  exit(1);
}

/*
 * Function Name:	wvSyslog
 * Description:		write log messages
 * Arguments:		log_t, log messages
 * Return Value:	none
 */
void
#if	__STDC__
wvSyslog(log_t type, const char * message, ...)
#else
wvSyslog(type, message, va_alist)
	log_t	type;
	char *	message;
	va_dcl
#endif	/* __STDC__ */
{
  va_list	ap;
#if	__STDC__
  va_start(ap, message);
#else
  va_start(ap);
#endif	/* __STDC__ */
  subsyslog(type, message, ap);
  va_end(ap);
}

/*
 * Function Name:	subsyslog
 * Description:		write log messages
 * Arguments:		log level, log messages
 * Return Value:	none
 */
static void
subsyslog(level, fmt, ap)
	log_t		level;
	const char *	fmt;
	va_list		ap;
{
  const static String	priority[] = {
    "unk",	/* dummy */
    "fatal",
    "emerg", "alert", "crit", "error", "warning", "notice", "info", "debug",
    "security"
  };

  if (level & LOG_SECURITY) {
    if (httpd->securityLog.fp != NULL) {
      fprintf(httpd->securityLog.fp, "[%s] ",
	      getTimeOfRFC822((time_t)0, TIMEZONE_LOCAL));
      (void)vfprintf(httpd->securityLog.fp, fmt, ap);
    }
    level &= 0x07;
  }

  if (level) {
    register FILE *	fout;

    fout = httpd->vhost == NULL ? stderr : httpd->vhost->errorLog.fp;
    fprintf(fout, "[%s] [%s] ",
	    getTimeOfRFC822((time_t)0, TIMEZONE_LOCAL), priority[level]);
    (void)vfprintf(fout, fmt, ap);
  }
}

