/*****************************************************************************
 *
 * FILE:	mime.c
 * DESCRIPTION:	register/get MIME types
 * DATE:	Sat, Sep 18 1999
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/cover.html
 * COPYRIGHT:	(C) 1999-2003 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *   Copyright (C) 1999-2003 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *   All rights reserved.
 *  
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions
 *   are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in
 *      the documentation and/or other materials provided with the
 *      distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY KOUICHI ABE (WALL) ``AS IS'' AND ANY
 *   EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL KOUICHI ABE (WALL) BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *   BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 *   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 *   OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 *   IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: mime.c,v 1.25.2.12 2003/03/18 16:06:06 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#if	HAVE_STDLIB_H
#include <stdlib.h>
#endif	/* HAVE_STDLIB_H */
#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#include "wyvern.h"
#include "functions.h"


/*****************************************************************************
 *
 *	definition of local macros and structures
 *
 *****************************************************************************/
typedef struct _Hash {
  String		key;
  String		value;
  struct _Hash *	next;
} Hash;


/*****************************************************************************
 *
 *	declaration of local variables
 *
 *****************************************************************************/
static Hash *	mimeTable = NULL;


/*****************************************************************************
 *
 *	declaration of local functions
 *
 *****************************************************************************/
static String	lookup(const String);


/*****************************************************************************
 *
 *	definition of functions
 *
 *****************************************************************************/

/*
 * Function Name:	mimeTypeEntry
 * Description:		register the suffix and the mimetype into hash table
 * Arguments:		suffix, mimetype
 * Return Value:	none
 */
void
mimeTypeEntry(suffix, type)
	const String	suffix;
	const String	type;
{
  Hash *	hp;	/* for current pointer	*/

  if (suffix == NULL) {
    return;	/* no suffix */
  }

  if (lookup(suffix) != NULL) {
    return;	/* already registered */
  }

  hp		= (Hash *)wvMalloc(sizeof(Hash));
  hp->key	= wvStrdup(suffix);
  hp->value	= wvStrdup(type);
  hp->next	= mimeTable;
  mimeTable	= hp;
}

/*
 * Function Name:	lookup
 * Description:		lookup mimetype with suffix
 * Arguments:		suffix
 * Return Value:	mimetype
 */
static String
lookup(suffix)
	const String	suffix;
{
  register Hash *	hp;

  for (hp = mimeTable; hp != NULL; hp = hp->next) {
    if (wvStrcmp(hp->key, suffix) == 0) {
      return hp->value;
    }
  }

  return NULL;
}

/*
 * Function Name:	defaultMimeTypeChecker
 * Description:		check mimetype of supplied file
 * Arguments:		message
 * Return Value:	successful: 0, otherwise: -1
 */
int
defaultMimeTypeChecker(msg)
	register Message *	msg;
{
  String	suffix;	/* suffix of file */

#if	DEBUG
  fputs("<defaultMimeTypeChecker>\n", stderr);
#endif	/* DEBUG */
  if (msg->file.mime != NULL) {
    return RET_OK;	/* already checked */
  }

  msg->file.mime	= msg->vhost->defaultType;
  suffix		= strrchr(msg->file.path, '.');
  if (suffix++ != NULL) {	/* no suffix */
    String	mtype = lookup(suffix);
    if (mtype != NULL) {
      msg->file.mime = mtype;
    }
  }

  return RET_OK;
}

