/*****************************************************************************
 *
 * FILE:	auth.h
 * DESCRIPTION:	authentication header
 * DATE:	Sat, Sep 18 1999
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/cover.html
 * COPYRIGHT:	(C) 1999-2003 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *  Copyright (C) 1999-2003 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *  
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in
 *      the documentation and/or other materials provided with the
 *      distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY KOUICHI ABE (WALL) ``AS IS'' AND ANY
 *   EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL KOUICHI ABE (WALL) BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *   BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 *   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 *   OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 *   IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: auth.h,v 1.12.6.3 2003/01/27 09:38:37 kouichi Exp $
 *
 *****************************************************************************/

#ifndef	_auth_h
#define	_auth_h

#include <sys/types.h>
#include <sys/stat.h>
#include "wpath.h"
#include "base64.h"


/*****************************************************************************
 *
 *	definition of macros and structure
 *
 *****************************************************************************/
#define	WP_NAME_SIZE		(10)
#define	WP_PASSWD_SIZE		(32+1)

typedef struct _File {
  String	path;	/* file path */
  time_t	mtime;	/* last modified time */
} File;

typedef struct _FileMgr {
  String		username;	/* username */
  File			waccess;	/* waccess */
  File			wpasswd;	/* wpasswd */
  struct _FileMgr *	next;
} FileMgr;

typedef struct _AuthList {
  String		name;
  struct _AuthList *	next;
} AuthList;

typedef struct _AuthDir {	/* manage authentication directory */
  String		dir;	/* restricted directory name	*/
  AuthList *		user;	/* allowed user name */
  AuthList *		addr;	/* allowed IP address or domain name */
  struct _AuthDir *	next;
} AuthDir;

typedef struct _Auth {	/* manage authentication */
  String		name;	/* user name (system is _wyvern_) */
  FileMgr *		file;
  AuthDir *		authDir;
  struct _Auth *	next;
} Auth;


/****************************************************************************/

extern FileMgr *	fileMgrLookup(const String);
extern int		authNew(const String, const String, const String);
extern void		authCleaner(void);
extern Auth *		authLookup(const String);
extern int		authUpdate(Auth *);
extern int		authChecker(Message *, AuthDir *,
				int (*)(Message *, const String, const String));

#endif	/* _auth_h */
