/*****************************************************************************
 *
 * FILE:	mod_log.c
 * DESCRIPTION:	wyvern another logging module
 * DATE:	Fri, Sep 15 2000
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/cover.html
 * COPYRIGHT:	(C) 2000-2003 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *  Copyright (C) 2000-2003 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *  
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in
 *      the documentation and/or other materials provided with the
 *      distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY KOUICHI ABE (WALL) ``AS IS'' AND ANY
 *   EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL KOUICHI ABE (WALL) BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *   BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 *   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 *   OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 *   IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: mod_log.c,v 1.11.2.16 2003/08/18 12:57:50 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#include "wyvern.h"
#include "functions.h"


/*****************************************************************************
 *
 *	definition of macros
 *
 *****************************************************************************/


/*****************************************************************************
 *
 *	declaration of local functions
 *
 *****************************************************************************/
static int	logWriter(Message *);
static void	logRegisterHooks(void);


/*****************************************************************************
 *
 *	definition of functions
 *
 *****************************************************************************/

/*
 * Function Name:	logWriter
 * Description:		write the log message in log file
 * Arguments:		message
 * Return Value:	successful: 0, otherwise: -1
 */
static int
logWriter(msg)
	register Message *	msg;
{
  String	client_name;

  client_name = msg->conn->client.name == NULL ? msg->conn->client.ip
					       : msg->conn->client.name;

  /* Client Error: 4xx	*/
  /* Server Error: 5xx	*/
  if (msg->response.status > STATUS_CODE_4XX) {
    if (msg->request.x_forwarded_for != NULL &&
	wvStrncasecmp(msg->request.x_forwarded_for, "unknown", 7) == 0) {
      wvSyslog(LOG_WARNING, "(%s) \"%s %s %s\" "
		"request forwarded from unknown host.\n",
		client_name,
		methodName[msg->request.method],
		msg->request.uri,
		httpVer[msg->request.http]);
    }
    else {
      log_t	log_type = LOG_ERR;

      if (msg->response.status == STATUS_FORBIDDEN ||
	  msg->response.status == STATUS_UNAUTHORIZED) {
	log_type |= LOG_SECURITY;
      }
      wvSyslog(log_type, "%s \"%s %s %s\" %s (%d)\n",
		client_name,
		methodName[msg->request.method],
		msg->request.uri,
		httpVer[msg->request.http],
		responseStatusTable[msg->response.status].phrase,
		responseStatusTable[msg->response.status].code);
    }
  }
  /* Informational: 1xx	*/
  /* Successful:    2xx	*/
  /* Redirection:   3xx	*/
  else {
    String	referer;
    String	user_agent;

    referer	= headerLookup(msg->request.headers, "REFERER");
    user_agent	= headerLookup(msg->request.headers, "USER_AGENT");

    if (msg->request.x_forwarded_for != NULL) {
      fprintf(msg->vhost->accessLog.fp, "%s(%s) ",
	      client_name, msg->request.x_forwarded_for);
    }
    else {
      fprintf(msg->vhost->accessLog.fp, "%s ", client_name);
    }
    fprintf(msg->vhost->accessLog.fp, "%s %s [%s] \"%s %s %s\" %d ",
	    msg->user.identName == NULL ? "-" : msg->user.identName,
	    msg->user.authName == NULL ? "-" : msg->user.authName,
	    getTimeOfRFC822(msg->accesstime, TIMEZONE_LOCAL),
	    methodName[msg->request.method],
	    msg->request.uri,
	    httpVer[msg->request.http],
	    responseStatusTable[msg->response.status].code);
    if (msg->response.status > STATUS_CODE_3XX) {
      fputs("- ", msg->vhost->accessLog.fp);
    }
    else {
      switch (msg->request.method) {
	case M_HEAD:
	  fputs("0 ", msg->vhost->accessLog.fp);
	  break;
	case M_GET:
	  fprintf(msg->vhost->accessLog.fp, "%ld ",
		  (long)msg->response.content_length);
	  break;
	case M_POST:
	  fprintf(msg->vhost->accessLog.fp, "%d ",
		  msg->request.content_length);
	  break;
	default:
	fputs("- ", msg->vhost->accessLog.fp);
	  break;
      }
    }
    fprintf(msg->vhost->accessLog.fp, "\"%s\" \"%s\"",
	    referer == NULL ? "-" : referer,
	    user_agent == NULL ? "-" : user_agent);
#if	USE_SSL
    if (msg->conn->sock.type == SOC_SECURE) {
      fprintf(msg->vhost->accessLog.fp, " %s %s",
	      SSL_get_version(msg->conn->sock.ssl),
	      SSL_get_cipher(msg->conn->sock.ssl));
    }
#endif	/* USE_SSL */
    fputs("\n", msg->vhost->accessLog.fp);
  }

  return RET_OK;
}


/*****************************************************************************
 *
 *	Module
 *
 *****************************************************************************/

static void
logRegisterHooks(void)
{
  wvHookLogger(logWriter, HOOK_MIDDLE);
}

Module	mod_log = {
  MODULE_MAGIC_NUMBER_MAJOR,	/* version */
  SAFETY_STANDARD,		/* safety */
  "mod_log",			/* name */
  "Log Module version 1.0",	/* info */
  "Kouichi ABE (WALL)",		/* author */
  NULL,				/* handle */
  NULL,				/* init */
  NULL,				/* finish */
  (void *)logRegisterHooks	/* register_handler */
};

Module *
getModuleInfo(void)
{
  return &mod_log;
}
