/*****************************************************************************
 *
 * FILE:	mod_md5.c
 * DESCRIPTION:	MD5 module for wyvern
 * DATE:	Tue, Aug 20 2002
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/cover.html
 * COPYRIGHT:	(C) 2002-2003 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *   Copyright (C) 2002-2003 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *   All rights reserved.
 *  
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions
 *   are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in
 *      the documentation and/or other materials provided with the
 *      distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY KOUICHI ABE (WALL) ``AS IS'' AND ANY
 *   EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL KOUICHI ABE (WALL) BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *   BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 *   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 *   OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 *   IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: mod_md5.c,v 1.1.2.10 2003/03/18 16:06:13 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_STDLIB_H
#include <stdlib.h>
#endif	/* HAVE_STDLIB_H */
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#ifndef	USE_SSL
#include <md5.h>
#endif	/* USE_SSL */
#include "wyvern.h"
#include "functions.h"


/*****************************************************************************
 *
 *	definition of macros and structures
 *
 *****************************************************************************/


/*****************************************************************************
 *
 *	declaration of local functions
 *
 *****************************************************************************/
static int	setMD5(Message *);
static void	md5RegisterHooks(void);


/*****************************************************************************
 *
 *	declaration of local variables
 *
 *****************************************************************************/


/*****************************************************************************
 *
 *	definition of functions
 *
 *****************************************************************************/

/*
 * Function Name:	setMD5
 * Description:		set the message digest of file entity
 * Arguments:		message
 * Return Value:	successful: 0, otherwise: -1
 */
static int
setMD5(msg)
	register Message *	msg;
{
  MD5_CTX	md5_ctx;
  unsigned char	md5[MD5_DIGEST_LENGTH];
  unsigned char	md5_hex[MD5_DIGEST_LENGTH*2+1];
  int		i;

  if (msg->response.status > STATUS_CODE_3XX ||
      msg->file.fd == -1) {	/* avoid potential bugs */
    return RET_OK;	/* no need */
  }
#if	DEBUG
  fputs("<setMD5>\n", stderr);
#endif	/* DEBUG */

  MD5Init(&md5_ctx);
  MD5Update(&md5_ctx, msg->file.data,
	    (size_t)msg->file.status.st_size);
  MD5Final(md5, &md5_ctx);

  memset(md5_hex, 0, sizeof(md5_hex));
  for (i = 0; i < MD5_DIGEST_LENGTH; i++) {
    (void)snprintf(md5_hex + i * 2, sizeof(md5_hex) - 1, "%02x", md5[i]);
  }
  (void)lseek(msg->file.fd, (off_t)0L, SEEK_SET);
  if (responseHeaderEntry(&msg->response.headers,
			  "Content-MD5", md5_hex) == NULL) {
    return wvNoMemoryError(msg);
  }
  return RET_OK;
}


/*****************************************************************************
 *
 *	Module
 *
 *****************************************************************************/

static void
md5RegisterHooks(void)
{
  wvHookFixups(setMD5, HOOK_LAST);
}

Module	mod_md5 = {
  MODULE_MAGIC_NUMBER_MAJOR,	/* version */
  SAFETY_STANDARD,		/* safety */
  "mod_md5",			/* name */
  "MD5 Module version 1.0",	/* info */
  "Kouichi ABE (WALL)",		/* author */
  NULL,				/* handle */
  NULL,				/* init */
  NULL,				/* finish */
  (void *)md5RegisterHooks	/* register_handler */
};

Module *
getModuleInfo(void)
{
  return &mod_md5;
}
