/*****************************************************************************
 *
 * FILE:	request.c
 * DESCRIPTION:	address the request URI from client
 * DATE:	Sun, Sep 26 1999
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/cover.html
 * COPYRIGHT:	(C) 1999-2003 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *   Copyright (C) 1999-2003 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *   All rights reserved.
 *  
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions
 *   are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in
 *      the documentation and/or other materials provided with the
 *      distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY KOUICHI ABE (WALL) ``AS IS'' AND ANY
 *   EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL KOUICHI ABE (WALL) BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *   BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 *   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 *   OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 *   IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: request.c,v 1.93.2.34 2003/03/18 16:06:06 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#if	HAVE_STDLIB_H
#include <stdlib.h>
#endif	/* HAVE_STDLIB_H */
#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#if	HAVE_FCNTL_H
#include <fcntl.h>
#endif	/* HAVE_FCNTL_H */
#if	HAVE_NETDB_H
#include <netdb.h>
#endif	/* HAVE_NETDB_H */
#include <sys/types.h>
#if	HAVE_SYS_SOCKET_H
#include <sys/socket.h>
#endif	/* HAVE_SYS_SOCKET_H */
#if	HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif	/* HAVE_NETINET_IN_H */
#if	HAVE_ARPA_INET_H
#include <arpa/inet.h>
#endif	/* HAVE_ARPA_INET_H */
#include <netinet/tcp.h>
#include <signal.h>
#include <sys/wait.h>
#if	TIME_WITH_SYS_TIME
#include <sys/time.h>
#include <time.h>
#else
#if	TM_IN_SYS_TIME
#include <sys/time.h>
#else
#include <time.h>
#endif	/* TM_IN_SYS_TIME */
#endif	/* TIME_WITH_SYS_TIME */
#include <errno.h>
#include "wyvern.h"
#include "wpath.h"
#include "functions.h"


/*****************************************************************************
 *
 *	definition of macros and structures
 *
 *****************************************************************************/
#define	SELECT_TIMEOUT	(5000)	/* msec. */


/*****************************************************************************
 *
 *	declaration of local variables
 *
 *****************************************************************************/


/*****************************************************************************
 *
 *	declaration of local functions
 *
 *****************************************************************************/
static int		polling(int);
static int		acceptRequest(Socket *);
static int		requestEngine(Thread *);
static int		responseEngine(Thread *);
static int		recvRequest(Thread *);
static Message *	mesgNew(Thread *);
static int		parseRequestLine(Message *, String);
static int		recvEntity(Thread *);
static String		decodeEscapedURI(String);
static int		isHex(char);


/*****************************************************************************
 *
 *	definition of functions
 *
 *****************************************************************************/

/*
 * Function Name:	httpdMain
 * Description:		address the request from each client
 * Arguments:		none
 * Return Value:	none
 */
void
httpdMain(void)
{
  int	timeout = SELECT_TIMEOUT;

  while (httpd->shutdown == FALSE) {
    int	ret;	/* return value of select() */

    /* check signals */
    if (httpd->trap.sigexit) {
      httpd->shutdown	= TRUE;	/* goto shutdown process... */
      timeout		= 0;
    }
    if (httpd->trap.sighup) {
      httpd->trap.sighup = FALSE;	/* ignore */
      closeLogFiles();
      openLogFiles();
    }

    /* request polling */
    ret = polling(timeout);
    switch (ret) {
      case 0:	/* timeout */
	if (threadSweep() == 0) {
	  timeout = TIMEOUT_INFINITY;
	}
	break;
      case -1:	/* error */
	if (errno == EINTR) {
	  continue;
	}
	die("killed by errno(%d) in select(2)\n", errno);
	break;
      default: {
	  register Socket *	sp;
	  register Thread *	tp;

	  /* servers */
	  for (sp = httpd->sockets; sp != NULL; sp = sp->next) {
	    if (sp->revents & POLL_IN) {
	      acceptRequest(sp);
	    }
	  }
	  /* workers */
	  for (tp = threadActives(); tp != NULL; tp = tp->next) {
	    if (tp->conn.sock.revents & POLL_IN) {
#if	USE_SSL
	      if (tp->conn.sock.ssl != NULL &&
		  !SSL_is_init_finished(tp->conn.sock.ssl)) {
		switch (sslAccept(tp)) {
		  case SSL_ERROR:
		    threadDetach(tp);
		  case SSL_RETRY:
		    continue;
		  case SSL_OK:
		    tp->status = TH_ACTIVE;
		  default:
		    break;
		}
	      }
#endif	/* USE_SSL */
	      switch (recvRequest(tp)) {
		case RET_OK:
		  requestEngine(tp);
		case RET_NG:
		  tp->conn.sock.events = POLL_OUT;
		  break;
		default:
		  (void)gettimeofday(&tp->time.current, (struct timezone *)0);
		  if (tp->time.current.tv_sec - tp->time.accepted.tv_sec > 5) {
#if	DEBUG
		    fprintf(stderr, "[TimeOut] cur: %ld, acpt: %ld\n",
			    tp->time.current.tv_sec, tp->time.accepted.tv_sec);
#endif	/* DEBUG */
		    threadDetach(tp);
		  }
		  break;
	      }
	    }
	    else if (tp->conn.sock.revents & POLL_OUT) {
	      responseEngine(tp);
	    }
	  }
	  /* reset timeout */
	  timeout = SELECT_TIMEOUT;
	}
	break;
    }
  }	/* while end */

  if (httpd->trap.sigexit) {
    wvSyslog(LOG_NOTICE, "terminated by sig#%d\n", httpd->trap.signo);
  }
  else {
    wvSyslog(LOG_INFO, "shutdown...\n");
  }
}

/*
 * Function Name:	polling
 * Description:		polling requests from clients
 * Arguments:		timeout (usec)
 * Return Value:	successful: 0, otherwise: -1
 */
static int
polling(timeout)
	int	timeout;
{
  register Socket *	sp;
  register Thread *	tp;
  fd_set		rfds;
  fd_set		wfds;
  int			maxfd;
  int			ret;

#if	DEBUG
  fprintf(stderr, "[polling] timeout = %d\n", timeout);
#endif	/* DEBUG */
  FD_ZERO(&rfds);
  FD_ZERO(&wfds);
  maxfd = httpd->maxsoc;

  /* servers */
  for (sp = httpd->sockets; sp != NULL; sp = sp->next) {
    FD_SET(sp->sd, &rfds);
  }
  /* workers */
  for (tp = threadActives(); tp != NULL; tp = tp->next) {
    if (tp->conn.sock.events & POLL_IN) {
      FD_SET(tp->conn.sock.sd, &rfds);
    }
    if (tp->conn.sock.events & POLL_OUT) {
      FD_SET(tp->conn.sock.sd, &wfds);
    }
    maxfd = MAX(maxfd, tp->conn.sock.sd);
  }

  /* polling */
  if (timeout == TIMEOUT_INFINITY) {
#if	DEBUG
    fputs("[polling] ...zzZ\n", stderr);
#endif	/* DEBUG */
    ret = select(maxfd + 1, &rfds, &wfds, NULL, NULL);
  }
  else {
    struct timeval	tv;
#if	DEBUG
    struct timeval	now;
    struct timeval	end;
    (void)gettimeofday(&now, (struct timezone *)0);
    fputs("[polling] select blocking ... ", stderr);
#endif	/* DEBUG */

    tv.tv_sec  = timeout / 1000;
    tv.tv_usec = (timeout % 1000) * 1000;
    ret = select(maxfd + 1, &rfds, &wfds, NULL, &tv);
#if	DEBUG
    (void)gettimeofday(&end, (struct timezone *)0);
    fprintf(stderr, "total = %ld.%ld\n",
	    end.tv_sec - now.tv_sec,
	    end.tv_sec - now.tv_sec == 0 ? end.tv_usec - now.tv_usec : 0);
#endif	/* DEBUG */
  }
  if (ret == -1) {
    return RET_NG;
  }

  /* servers */
  for (sp = httpd->sockets; sp != NULL; sp = sp->next) {
    sp->revents = 0;
    if (FD_ISSET(sp->sd, &rfds)) {
      sp->revents |= POLL_IN;
#if	DEBUG
      fprintf(stderr, "[polling] server(%d) acceptable\n", sp->sd);
#endif	/* DEBUG */
    }
  }
  /* workers */
  for (tp = threadActives(); tp != NULL; tp = tp->next) {
    tp->conn.sock.revents = 0;
    if (FD_ISSET(tp->conn.sock.sd, &rfds)) {
      tp->conn.sock.revents |= POLL_IN;
#if	DEBUG
      fprintf(stderr, "[polling] worker(%d) readable, errno=%d\n",
	      tp->conn.sock.sd, errno);
#endif	/* DEBUG */
    }
    if (FD_ISSET(tp->conn.sock.sd, &wfds)) {
      tp->conn.sock.revents |= POLL_OUT;
#if	DEBUG
      fprintf(stderr, "[polling] worker(%d) writable, errno=%d\n",
	      tp->conn.sock.sd, errno);
#endif	/* DEBUG */
    }
  }

  return ret;
}

/*
 * Function Name:	acceptRequest
 * Description:		accept new request
 * Arguments:		socket pointer
 * Return Value:	successful: 0, otherwise: -1
 */
static int
acceptRequest(sp)
	Socket *	sp;
{
  struct sockaddr_storage	ss;
  struct sockaddr *		cliaddr = (struct sockaddr *)&ss;
  socklen_t			fromlen = sizeof(ss);
  char				hbuf[NI_MAXHOST];
  char				sbuf[NI_MAXSERV];
  int				sd;	/* new socket descriptor */
  int				on;	/* flag to set socket option */
  Thread *			thd;

  /* new request from client */
  sd = accept(sp->sd, cliaddr, &fromlen);
  if (sd == -1) {
    return RET_NG;
  }

  /* allow reuse of local address (error is ignored) */
  on = 1;
  (void)setsockopt(sd, SOL_SOCKET, SO_REUSEADDR, (void *)&on, sizeof(on));

  /* change new socket descriptor to non-blocking mode */
  if (setNonBlocking(sd) == RET_NG) {
    wvSyslog(LOG_ERR, "failed to change non-blocking mode\n");
    wvClose(sd);
    return RET_NG;
  }

  if (getnameinfo(cliaddr, fromlen, hbuf, sizeof(hbuf), sbuf, sizeof(sbuf),
		  NI_NUMERICHOST|NI_NUMERICSERV) != 0) {
    wvSyslog(LOG_ERR, "failed to get info of remote host\n");
    wvClose(sd);
    return RET_NG;
  }

  thd = threadNew(sd);	/* yield new thread */
  if (thd == NULL) {	/* thread queue is full */
    wvClose(sd);
    wvSyslog(LOG_ERR, "No more threads\n");
    return RET_NG;
  }
  /* timestamp */
  (void)gettimeofday(&thd->time.accepted, (struct timezone *)0);
  thd->time.last = thd->time.current = thd->time.accepted;
  /* socket */
  thd->conn.sock.type	= sp->type;
  thd->conn.sock.port	= sp->port;
  thd->conn.sock.ainfo	= sp->ainfo;
  /* set client information */
  thd->conn.client.port	= wvStrdup(sbuf);
  thd->conn.client.ip	= wvStrdup(hbuf);
  thd->conn.client.name	= NULL;
  thd->conn.client.addr	= ss;
  /**/
  thd->mqueue.conn	= &thd->conn;
#if	DEBUG
  fprintf(stderr, "[client]: port = %s\n", thd->conn.client.port);
#endif	/* DEBUG */

#if	USE_SSL
  if (sp->type == SOC_SECURE) {
    thd->conn.sock.ssl = SSL_new(sp->ssl_ctx);
    if (thd->conn.sock.ssl == NULL) {
      return RET_NG;
    }
    SSL_clear(thd->conn.sock.ssl);
    SSL_set_fd(thd->conn.sock.ssl, thd->conn.sock.sd);	/* bind decriptor */
    SSL_set_read_ahead(thd->conn.sock.ssl, 1);
    SSL_set_accept_state(thd->conn.sock.ssl);
  }
#endif	/* USE_SSL */

  return RET_OK;
}

/*
 * Function Name:	requestEngine
 * Description:		transact requests
 * Arguments:		thread
 * Return Value:	successful: 0, otherwise: -1
 */
static int
requestEngine(thd)
	register Thread *	thd;
{
  register Message *	mp;

  for (mp = &thd->mqueue; mp != NULL; mp = mp->next) {
    mp->status = MSG_WRITING_READY;

    /* parse headers */
    if (wvRunHeaderParser(mp) == RET_NG) {
      continue;
    }

    /* access checking */
    if (wvRunAccessChecker(mp) == RET_NG) {
      if (thd->conn.client.name == NULL) {
	wvSyslog(LOG_SECURITY, "access denied from %s\n", thd->conn.client.ip);
      }
      else {
	wvSyslog(LOG_SECURITY, "access denied from %s(%s)\n",
				thd->conn.client.name, thd->conn.client.ip);
      }
      threadDetach(thd);
      return RET_NG;
    }

    if (wvRunTranslateURI(mp) == RET_NG ||	/* uri translating */
	wvRunCheckUserID(mp) == RET_NG ||
	wvRunAuthChecker(mp) == RET_NG ||	/* authenticating */
	wvRunTypeChecker(mp) == RET_NG) {	/* mimetype checking */
      continue;
    }

    if (mp->request.method == M_HEAD) {
      mp->response.status = STATUS_REQUEST_OK;
      continue;
    }

    /* prepare to handle */
    if (wvRunFixups(mp, HOOK_FIRST) == RET_OK) {
      /* request handling */
      switch (wvRunHandler(mp)) {
	case RET_NG:
	  continue;
	case RET_WAIT:
#if	DEBUG
	  fputs("[requestEngine] cgi running\n", stderr);
#endif	/* DEBUG */
	  break;
	default:
	  wvRunFixups(mp, HOOK_MIDDLE);	/* prepare for filtering */
	  wvRunFilter(mp);
	  errorDocChecker(mp);
	  fileMapping(mp);		/* mapping files into memory */
	  wvRunFixups(mp, HOOK_LAST);	/* preparing to response */
	  break;
      }
    }
  }
  (void)gettimeofday(&thd->time.current, (struct timezone *)0);
#if	DEBUG
  fprintf(stderr, "[requestEngine] working time = %ld.%ld\n",
	  thd->time.current.tv_sec - thd->time.last.tv_sec,
	  thd->time.current.tv_sec - thd->time.last.tv_sec == 0 ?
	  thd->time.current.tv_usec - thd->time.last.tv_usec : 0);
#endif	/* DEBUG */
  thd->time.last = thd->time.current;

  return RET_OK;
}

/*
 * Function Name:	responseEngine
 * Description:		transact requests for response
 * Arguments:		thread
 * Return Value:	successful: 0, otherwise: -1
 */
static int
responseEngine(thd)
	register Thread *	thd;
{
  register Message *	mp;
  bool_t		ka_flag = FALSE;

  for (mp = &thd->mqueue; mp != NULL; mp = mp->next) {
    switch (mp->status) {
      case MSG_PROCESSING:		/* for CGI */
	mp->status = MSG_WRITING_READY;
	wvRunFixups(mp, HOOK_MIDDLE);	/* prepare for filtering */
	wvRunFilter(mp);

      case MSG_ERROR:			/* for error response */
	mp->status = MSG_WRITING_READY;
	errorDocChecker(mp);
	fileMapping(mp);		/* mapping files into memory */
	wvRunFixups(mp, HOOK_LAST);	/* preparing to response */

      case MSG_WRITING:;
      case MSG_WRITING_READY:
	responseHandler(thd);
	if (mp->status == MSG_WRITING) {
	  break;
	}

      case MSG_LOGGING:;
	wvRunLogger(mp);

      case MSG_CLOSING:
	if (thd->conn.keepalive.status == KA_CONTINUE &&
	    thd->conn.keepalive.max > 0 &&
	    thd->time.current.tv_sec - thd->time.accepted.tv_sec
						< thd->conn.keepalive.timeout) {
	  ka_flag = TRUE;
	  thd->conn.sock.events = POLL_IN;
#if	DEBUG
	  fprintf(stderr,
		  "[responseEngine] keep-alive continue: rest %d time %ld\n",
		  thd->conn.keepalive.max,
		  thd->conn.keepalive.timeout
		  - (thd->time.current.tv_sec - thd->time.accepted.tv_sec));
#endif	/* DEBUG */
	  break;
	}
	threadDetach(thd);
#if	DEBUG
	fputs("[responseEngine] closing\n", stderr);
#endif	/* DEBUG */
	return RET_OK;

      case MSG_IDLING:	/* for non-waiting CGI such as NPH-script */
      case MSG_WAITING:	/* for CGI etc */
	(void)gettimeofday(&thd->time.current, (struct timezone *)0);
	if (thd->time.current.tv_sec - thd->time.accepted.tv_sec >
							mp->vhost->timeout) {
	  kill(mp->cgi.pid, SIGTERM);
	  mp->response.status = STATUS_REQUEST_TIMEOUT;
	  mp->status	      = MSG_WRITING_READY;
	}
	break;

      default:
	break;
    }
  }
  (void)gettimeofday(&thd->time.current, (struct timezone *)0);
  /* initialize thread */
  if (ka_flag) {
    threadClear(thd);
  }
#if	DEBUG
  fprintf(stderr, "[responsetEngine] working time = %ld.%ld\n",
	  thd->time.current.tv_sec - thd->time.last.tv_sec,
	  thd->time.current.tv_sec - thd->time.last.tv_sec == 0 ?
	  thd->time.current.tv_usec - thd->time.last.tv_usec : 0);
#endif	/* DEBUG */
  thd->time.last = thd->time.current;

  return RET_OK;
}

/*
 * Function Name:	recvRequestLine
 * Description:		receive the request from a client
 * Arguments:		thread
 * Return Value:	successful: 0, otherwise: -1
 */
static int
recvRequest(thd)
	register Thread *	thd;
{
  int			n;
  char			buf[BUF_SIZE];
  register Message *	msg = &thd->mqueue;
  int			ret = RET_CONT;

  /* skip first white lines (for old buggy browser) */
  errno = 0;
  memset(buf, 0, sizeof(buf));
  while ((n = readline(thd, buf, (size_t)sizeof(buf))) > 0) {
    switch (msg->status) {
      case MSG_NONE:
	if (parseRequestLine(msg, buf) == RET_NG) {
	  msg->status = MSG_ERROR;
	  return RET_NG;
	}
	msg->status = MSG_READING;
	break;
      case MSG_READING:
	if (buf[0] == CR && buf[1] == LF) {	/* check CRLF */
	  time(&msg->accesstime);
	  msg->status = MSG_READY;
	  msg->conn   = &thd->conn;
#if	DEBUG
	  fprintf(stderr, "[recvRequest] sock=%d type=%d\n",
		  msg->conn->sock.sd, msg->conn->sock.type);
#endif	/* DEBUG */
	  memset(buf, 0, sizeof(buf));
	  ret = RET_OK;
	  /* post request */
	  if (msg->request.method == M_POST) {
	    thd->msg = msg;
	    return recvEntity(thd);
	  }
	  continue;
	}
	/* set request header fields */
	if (requestHeaderEntry(&msg->request.headers, buf) == NULL) {
	  wvSyslog(LOG_ERR, "failed to register request header: %s\n", buf);
	  msg->response.status	= STATUS_INTERNAL_SERVER_ERROR;
	  msg->status		= MSG_ERROR;
	  return RET_NG;
	}
	break;
      case MSG_READY:
	/* pipelining */
	msg->next = mesgNew(thd);
	if (msg->next == NULL) {
	  return wvNoMemoryError(msg);
	}
	msg = msg->next;
#if	DEBUG
	fprintf(stderr, "[PIPELINE] request: %s", buf);
#endif	/* DEBUG */
	if (parseRequestLine(msg, buf) == RET_NG) {
	  return RET_NG;
	}
	msg->status = MSG_READING;
	break;
      default:
	break;
    }
    memset(buf, 0, sizeof(buf));
  }

  return ret;
}

/*
 * Function Name:	mesgNew
 * Description:		create new message object
 * Arguments:		thread
 * Return Value:	successful: Message, otherwise: NULL
 */
static Message *
mesgNew(thd)
	register Thread *	thd;
{
  Message *	new;

  new = (Message *)calloc(1, sizeof(Message));
  if (new == NULL) {
    return NULL;
  }
  new->status	= MSG_NONE;
  new->vhost	= thd->mqueue.vhost;
  new->conn	= &thd->conn;
  new->next	= NULL;
  new->request.method  = M_NONE;
  new->request.http    = HTTP_10;
  new->response.status = STATUS_NONE;
  return new;
}

/*
 * Function Name:	parseRequestLine
 * Description:		parse the request line from a client
 * Arguments:		thread, buffer
 * Return Value:	successful: 0, otherwise: -1
 */
static int
parseRequestLine(msg, line)
	register Message *	msg;
	register String		line;
{
  int		n;
  String *	ap;
  String	args[3];
  String	p;

  /* divide the request line into the method, URI and protocol */
  for (n = 0, p = wvChomp(line), ap = args; (*ap = strsep(&p, " ")) != NULL; ) {
    if (**ap != EOL) {
      n++;
      if (++ap >= &args[3]) {
	break;
      }
    }
  }

  /* invalid request-line */
  if (n < 2) {
    msg->response.status = STATUS_BAD_REQUEST;
    return RET_NG;
  }
#define	HTTP_METHOD	(0)
#define	HTTP_URI	(1)
#define	HTTP_VERSION	(2)
  switch (n) {
    case 3:	/* set HTTP PROTOCOL */
      if (wvStrcasecmp(args[HTTP_VERSION], "HTTP/0.9") == 0) {
	msg->request.http = HTTP_09;
      }
      else if (wvStrcasecmp(args[HTTP_VERSION], "HTTP/1.0") == 0) {
	msg->request.http = HTTP_10;
      }
      else if (wvStrcasecmp(args[HTTP_VERSION], "HTTP/1.1") == 0) {
	msg->request.http = HTTP_11;
	if (msg->conn->keepalive.max > 0) {
	  msg->conn->keepalive.max--;
	  msg->conn->keepalive.status = KA_CONTINUE;
	}
      }
      else {
	wvFree(httpVer[HTTP_00]);
	httpVer[HTTP_00]	= wvStrdup(args[HTTP_VERSION]);
	msg->request.http	= HTTP_00;
	msg->response.status	= STATUS_HTTP_VERSION_NOT_SUPPORTED;
      }
    case 2:	/* set URI */
      wvStrncpy(msg->request.uri, args[HTTP_URI], sizeof(msg->request.uri));
      if (args[HTTP_URI][0] != NULL &&
	  strlen(args[HTTP_URI]) >= sizeof(msg->request.uri) - 1 &&
	  msg->response.status == STATUS_NONE) {
	msg->response.status = STATUS_REQUEST_URI_TOO_LONG;
      }
    case 1:	/* set METHOD */
      if (wvStrcasecmp(args[HTTP_METHOD], "HEAD") == 0) {
	msg->request.method = M_HEAD;
      }
      else if (wvStrcasecmp(args[HTTP_METHOD], "GET") == 0) {
	msg->request.method = M_GET;
      }
      else if (wvStrcasecmp(args[HTTP_METHOD], "POST") == 0) {
	msg->request.method = M_POST;
      }
      else {
#if	1
	if (msg->response.status == STATUS_NONE) {
	  wvFree(methodName[M_UNKNOWN]);
	  methodName[M_UNKNOWN]	= wvStrdup(args[HTTP_METHOD]);
	  msg->request.method	= M_UNKNOWN;
	  msg->response.status	= STATUS_METHOD_NOT_ALLOWED;
	  (void)responseHeaderEntry(&msg->response.headers,
				    "Allow", ALLOW_METHODS);
	}
#else
	msg->response.status = STATUS_NOT_IMPLEMENTED;
#endif
	return RET_NG;
      }
      break;
    default:
      msg->response.status = STATUS_BAD_REQUEST;
      return RET_NG;
  }

  return msg->response.status > STATUS_CODE_4XX ? RET_NG : RET_OK;
}

/*
 * Function Name:	recvEntity
 * Description:		receive the request entity from a client
 * Arguments:		thread
 * Return Value:	successful: 0, otherwise: -1
 */
static int
recvEntity(thd)
	register Thread *	thd;
{
  String	content_length;
  char		buf[BUF_SIZE];
  int		len;
  int		n;

  /* conteent-length */
  content_length = headerLookup(thd->msg->request.headers, "CONTENT_LENGTH");
  if (content_length == NULL) {
    thd->msg->response.status = STATUS_LENGTH_REQUIRED;
    return RET_NG;
  }
  thd->msg->request.content_length = atoi(content_length);
  if (thd->msg->request.content_length > MAX_FILE_SIZE) {
    thd->msg->response.status = STATUS_REQUEST_ENTITY_TOO_LARGE;
    return RET_NG;
  }

  /* create temporary filename */
  if (asprintf(&thd->msg->request.post_tmp,
	       "%s/wvXXXXXX", httpd->tmpDir) == -1) {
    goto error;
  }

  /* open temporary file */
  thd->msg->request.post_fd = mkstemp(thd->msg->request.post_tmp);
  if (thd->msg->request.post_fd == -1) {
    goto error;
  }
  /* the file will be removed when it's closed. */
  unlink(thd->msg->request.post_tmp);

  /* read data from client */
  for (n = 0, len = 0; len < thd->msg->request.content_length; len += n) {
    memset(buf, 0, sizeof(buf));
    if ((n = nread(thd, buf, sizeof(buf))) > 0) {
      if (write(thd->msg->request.post_fd, buf, (size_t)n) != n) {
	goto error;
      }
    }
  }
  /* rewind file pointer */
  lseek(thd->msg->request.post_fd, (off_t)0, SEEK_SET);
  return RET_OK;

error:
  wvSyslog(LOG_ERR, "failed to receive the request entity\n");
  thd->msg->response.status = STATUS_INTERNAL_SERVER_ERROR;
  return RET_NG;
}

/*
 * Function Name:	parseURI
 * Description:		parse request uri
 * Arguments:		message
 * Return Value:	successful: 0, otherwise: -1
 */
int
parseURI(msg)
	register Message *	msg;
{
  register String	urip;
  register String	p;

#if	DEBUG
  fputs("<parseURI>\n", stderr);
#endif	/* DEBUG */
  /* set a query_string */
  for (urip = msg->request.uri; *urip != EOL && *urip != '?'; urip++) {
    ;	/* nothing to do */
  }
  if (*urip == '?') {
    *urip++ = EOL;	/* overwrite '?' to EOL */
    msg->request.query_string = wvStrdup(wvChomp(urip));
  }

  /* parse a percent-encoded URI (URI length is short after decoding) */
  decodeEscapedURI(msg->request.uri);

  /* Forbid a request URI with relative path */
  urip = (String)strstr(msg->request.uri, "/..");
  if (urip != NULL) {
    if (urip[3] == '/' || urip[3] == EOL) {
      msg->response.status = STATUS_FORBIDDEN;
      return RET_NG;
    }
  }

  /* authentication files */
  p = strrchr(msg->request.uri, '/');
  if (p++ != NULL) {
    if (wvStrcmp(p, USER_ACCESS) == 0 || wvStrcmp(p, USER_PASSWD) == 0) {
      msg->response.status = STATUS_FORBIDDEN;
      return RET_NG;
    }
  }

  /* switching virtual host */
  return vhostSwitch(msg);
}

/*
 * Function Name:	decodeEscapedURI
 * Description:		decode an escaped character
 * Arguments:		string
 * Return Value:	decoded string
 * Note:		destroy original string
 * Reference:		RFC 2396
 *			Uniform Resource Identifiers (URI): Generic Syntax
 */
static String
decodeEscapedURI(s)
	String	s;
{
  char			tmp[MAX_URI_LEN];
  register String	os  = tmp;
  register String	ret = s;
  char			c, c1;

  /* copy original request URI into working buffer */
  memset(tmp, 0, sizeof(tmp));
  memcpy(tmp, s, sizeof(tmp));

  while ((c = *os) != EOL) {
    /* escaped URI */
    if (c == '%') {
      os++;
      if ((c = *os++) && (c1 = *os++)) {
	if (isHex(c) == 0 && isHex(c1) == 0) {
	  *s =	((c  >= 'A' ? ((c  & 0xdf) - 'A') + 10 : c  - '0') * 16) +
		((c1 >= 'A' ? ((c1 & 0xdf) - 'A') + 10 : c1 - '0'));
	  c = *s;
	  os--;	/* back one word */
	}
	else {	/* no hex digit */
	  *s++	= '%';
	  *s++	= c;
	  *s++	= c1;
	  continue;
	}
      }
      else {
        return ret;
      }
    }
    /* normal characters */
    *s++ = c;
    os++;
  }
  *s = EOL;
  return ret;
}

static int
isHex(c)
	register char	c;
{
  if ((c >= '0' && c <= '9') ||
      (c >= 'a' && c <= 'f') || (c >= 'A' && c <= 'F')) {
    return RET_OK;
  }
  return RET_NG;
}
