/*****************************************************************************
 *
 * FILE:	response.c
 * DESCRIPTION:	response the data to the client
 * DATE:	Sun, Oct  3 1999
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/cover.html
 * COPYRIGHT:	(C) 1999-2003 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *   Copyright (C) 1999-2003 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *   All rights reserved.
 *  
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions
 *   are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in
 *      the documentation and/or other materials provided with the
 *      distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY KOUICHI ABE (WALL) ``AS IS'' AND ANY
 *   EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL KOUICHI ABE (WALL) BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *   BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 *   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 *   OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 *   IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: response.c,v 1.72.2.18 2003/08/18 12:57:45 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#if	HAVE_STDLIB_H
#include <stdlib.h>
#endif	/* HAVE_STDLIB_H */
#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#include <time.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <time.h>
#include "wyvern.h"
#include "functions.h"


/*****************************************************************************
 *
 *	definition of macros and structures
 *
 *****************************************************************************/
#define	TIME_LIMIT	(5)	/* send (TIME_LIMIT * RESPONSE_BUFSIZE) bytes */
				/* at once */


/*****************************************************************************
 *
 *	declaration of local functions
 *
 *****************************************************************************/
static void	sendResponse(Thread *);
static void	sendStatusLine(Thread *);
static void	sendResponseHeader(Thread *);
static void	sendResponseEntity(Thread *);
static void	sendingEntity(Thread *);
static void	responseDone(Thread *);
#if	0
static void	setTransferEncoding(Thread *);
#endif


/*****************************************************************************
 *
 *	declaration of local variables
 *
 *****************************************************************************/
static int	suspend;


/*****************************************************************************
 *
 *	definition of functions
 *
 *****************************************************************************/

/*
 * Function Name:	responseHandler
 * Description:		transact response
 * Arguments:		thread
 * Return Value:	successful: 0, otherwise: -1
 */
void
responseHandler(thd)
	register Thread *	thd;
{
  for (thd->msg = &thd->mqueue; thd->msg != NULL; thd->msg = thd->msg->next) {
    switch (thd->msg->status) {
      case MSG_WRITING_READY:
	sendResponse(thd);
	break;
      case MSG_WRITING:
	sendingEntity(thd);
	break;
      default:
	thd->msg->status = MSG_CLOSING;
	break;
    }
  }
}

/*
 * Function Name:	sendResponse
 * Description:		send the response to a client
 * Arguments:		thread
 * Return Value:	none
 */
void
sendResponse(thd)
	register Thread *	thd;
{
  /* send status line */
  sendStatusLine(thd);

  /* send header fields and entity */
  switch (thd->msg->request.method) {
    case M_GET:
    case M_POST:
      sendResponseHeader(thd);
      sendResponseEntity(thd);
      break;
    case M_HEAD:
    default:
      sendResponseHeader(thd);
      responseDone(thd);
      break;
  }
}


/*
 * Function Name:	sendStatusLine
 * Description:		send HTTP status line for request
 * Arguments:		thread
 * Return Value:	none
 */
static void
sendStatusLine(thd)
	register Thread *	thd;
{
  if (thd->msg->cgi.nph == FALSE) {
    if (thd->msg->cgi.status != NULL) {
      iosPrintf(thd, "%s %s",
  		httpVer[thd->msg->request.http], thd->msg->cgi.status);
    }
    else {
      iosPrintf(thd, "%s %d %s" CRLF,
		      httpVer[thd->msg->request.http],
		      responseStatusTable[thd->msg->response.status].code,
		      responseStatusTable[thd->msg->response.status].phrase);
    }
  }
}

/*
 * Function Name:	sendResponseHeader
 * Description:		send the response header to a client
 * Arguments:		thread
 * Return Value:	none
 */
static void
sendResponseHeader(thd)
	register Thread *	thd;
{
  register Headers *	hp;

  thd->msg->status = MSG_WRITING;

  /* non-parsed header */
  if (thd->msg->cgi.nph == TRUE) {
    return;
  }

  /* send standard response header fields */
  iosPrintf(thd, "Date: %s" CRLF
		 "Server: %s" CRLF,
		 getTimeOfRFC822((time_t)0, TIMEZONE_GMT),
		 SERVER_VERSION);
 
  /**/
  for (hp = thd->msg->response.headers; hp != NULL; hp = hp->next) {
    iosPrintf(thd, "%s: %s" CRLF, hp->name, hp->value);
  }

  /* 4xx/5xx */
  if (thd->msg->response.status > STATUS_CODE_4XX) {
    iosPuts(thd, "Content-Type: text/html" CRLF
		 "Connection: close" CRLF
		 CRLF);
    thd->conn.keepalive.status = KA_CLOSE;
    return;
  }

  /* 3xx */
  if (thd->msg->response.status > STATUS_CODE_3XX) {
    iosPuts(thd, "Connection: close" CRLF CRLF);
    thd->conn.keepalive.status = KA_CLOSE;
    return;
  }

  /* 2xx */
  if (thd->msg->file.mime != NULL) {
    iosPrintf(thd, "Content-Type: %s" CRLF, thd->msg->file.mime);
  }
  if (thd->msg->response.content_length > 0) {
    iosPrintf(thd, "Content-Length: %d" CRLF,
	      thd->msg->response.content_length);
  }
  if (thd->msg->file.status.st_mtime > 0) {
    iosPrintf(thd, "Last-Modified: %s" CRLF,
	      getTimeOfRFC822(thd->msg->file.status.st_mtime, TIMEZONE_GMT));
  }

  /** header for Keep-Alive mode */
  if (thd->conn.keepalive.status == KA_CONTINUE &&
      thd->conn.keepalive.max > 0) {
#if	BUGGY_KEEP_ALIVE
    if (thd->msg->request.http == HTTP_10) {
      iosPrintf(thd, "Connection: Keep-Alive" CRLF
		     "Keep-Alive: timeout=%d, max=%d" CRLF,
		      thd->conn.keepalive.timeout
		      - (thd->time.current.tv_sec - thd->time.accepted.tv_sec),
		      thd->conn.keepalive.max);
    }
#endif	/* BUGGY_KEEP_ALIVE */
  }
  else {
    iosPuts(thd, "Connection: close" CRLF);
    thd->conn.keepalive.status = KA_CLOSE;
  }

  iosPuts(thd, CRLF);	/* response header field end */
}

/*
 * Function Name:	sendResponseEntity
 * Description:		send the response entity to a client
 * Arguments:		thread
 * Return Value:	none
 */
static void
sendResponseEntity(thd)
	register Thread *	thd;
{
  /* send the entity of error	*/
  if (thd->msg->response.status > STATUS_CODE_4XX) {
    if (thd->msg->file.fd == -1) {
      errorMessageSender(thd);
      responseDone(thd);
    }
    else {
      sendingEntity(thd);
    }
  }
  else if (thd->msg->response.status > STATUS_CODE_3XX) {
    responseDone(thd);	/* send no entity! */
  }
  else {
    sendingEntity(thd);
  }
}

void
responseTimeout(sig)
	int	sig;
{
  suspend = 1;
}

/*
 * Function Name:	sendingEntity
 * Description:		send the response entity to a client
 * Arguments:		thread
 * Return Value:	none
 */
void
sendingEntity(thd)
	register Thread *	thd;
{
  struct timeval	now;
  struct timeval	active;
  struct itimerval	timer;
  size_t		len;
  size_t		size;

  (void)gettimeofday(&now, (struct timezone *)0);
  active.tv_sec	 = now.tv_sec - thd->time.last.tv_sec;
  active.tv_usec = active.tv_sec > 0 ? 0 : now.tv_usec - thd->time.last.tv_usec;

  timer.it_interval.tv_sec	= 0;
  timer.it_interval.tv_usec	= 0;
  timer.it_value.tv_sec		= active.tv_sec > 0 ? active.tv_sec : 1;
  timer.it_value.tv_usec	= active.tv_usec;
  setitimer(ITIMER_VIRTUAL, &timer, NULL);
  suspend = 0;
#if	DEBUG
  fprintf(stderr, "\t[sending] timer = %ld.%ld\t",
	  timer.it_value.tv_sec, timer.it_value.tv_usec);
#endif	/* DEBUG */
  /* sending data */
  for (len = 0, size = MIN(RESPONSE_BUFSIZE, thd->msg->file.rest);
       thd->msg->file.rest > 0;
       thd->msg->file.curp += len, thd->msg->file.rest -= len,
       size = MIN(RESPONSE_BUFSIZE, thd->msg->file.rest)) {
    len = iosWrite(thd, thd->msg->file.curp, size);
    if (suspend == 1) {
#if	DEBUG
      fputs("timeout\n", stderr);
#endif	/* DEBUG */
      return;
    }
  }
  timer.it_interval.tv_sec	= 0;
  timer.it_interval.tv_usec	= 0;
  timer.it_value.tv_sec		= 0;
  timer.it_value.tv_usec	= 0;
  setitimer(ITIMER_VIRTUAL, &timer, NULL);
#if	DEBUG
  fputs("ok\n", stderr);
#endif	/* DEBUG */

  responseDone(thd);
}

/*
 * Function Name:	responseDone
 * Description:		finish to send the response entity to a client
 * Arguments:		thread
 * Return Value:	none
 */
static void
responseDone(thd)
	register Thread *	thd;
{
  iosFlush(thd);
  thd->msg->status = MSG_CLOSING;
  (void)munmap(thd->msg->file.data, (size_t)thd->msg->file.status.st_size);
  wvClose(thd->msg->file.fd);
  thd->msg->file.fd = -1;
#if	DEBUG
  if (thd->conn.keepalive.status == KA_CONTINUE) {
    fprintf(stderr, "[worker(%d)]: keep-alive continued...\n",
	    thd->conn.sock.sd);
  }
  else {
    fprintf(stderr, "[(%s)]: sending entity ... done\n", thd->conn.client.port);
  }
#endif	/* DEBUG */
}


#if	0
/*
 * Function Name:	setTransferEncoding
 * Description:		set the Transfer-Encoding according to TE field
 * Arguments:		thread
 * Return Value:	none
 */
static void
setTransferEncoding(thd)
	Thread *	thd;
{
  if (thd->request.te == NULL) {
    thd->response.transfer_encoding = wvStrdup("chunked");
    iosPrintf(thd, "Transfer-Encoding: %s" CRLF,
		   thd->response.transfer_encoding);
  }
  else {
    /*
     * Not implemented
     */
  }
}
#endif

