/*****************************************************************************
 *
 * FILE:	signal.c
 * DESCRIPTION:	signal handler
 * DATE:	Mon, Nov 15 1999
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/cover.html
 * COPYRIGHT:	(C) 1999-2003 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *   Copyright (C) 1999-2003 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *   All rights reserved.
 *  
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions
 *   are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in
 *      the documentation and/or other materials provided with the
 *      distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY KOUICHI ABE (WALL) ``AS IS'' AND ANY
 *   EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL KOUICHI ABE (WALL) BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *   BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 *   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 *   OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 *   IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: signal.c,v 1.22.2.14 2003/03/18 16:06:06 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#include <time.h>
#include <signal.h>
#if	SIGTSTP
#if	HAVE_SYS_FILE_H
#include <sys/file.h>
#endif	/* HAVE_SYS_FILE_H */
#if	HAVE_SYS_IOCTL_H
#include <sys/ioctl.h>
#endif	/* HAVE_SYS_IOCTL_H */
#endif	/* SIGTSTP */
#if	HAVE_SYS_WAIT_H
#include <sys/wait.h>
#endif	/* HAVE_SYS_WAIT_H */
#include "wyvern.h"
#include "functions.h"


/*****************************************************************************
 *
 *	declaration of local functions
 *
 *****************************************************************************/
static void	trap(int);
static void	reapChild(void);
#if	DEBUG
static void	debugSignal(void);
#endif	/* DEBUG */


/*****************************************************************************
 *
 *	declaration of local variables
 *
 *****************************************************************************/


/*****************************************************************************
 *
 *	definition of functions
 *
 *****************************************************************************/

/*
 * Function Name:	signalTrap
 * Description:		set the signal handler
 * Arguments:		none
 * Return Value:	none
 */
void
signalTrap(void)
{
  struct sigaction	act;

  /* quit */
  sigemptyset(&act.sa_mask);
  act.sa_flags	 = 0;
  act.sa_handler = (void *)trap;
  sigaction(SIGINT,  &act, NULL);
  sigaction(SIGQUIT, &act, NULL);
  sigaction(SIGTERM, &act, NULL);
  /* bug */
  sigaction(SIGBUS,  &act, NULL);
  sigaction(SIGSEGV, &act, NULL);
  /* restart */
  act.sa_handler = (void *)trap;
  sigaction(SIGHUP,  &act, NULL);

  /* response timer */
  act.sa_flags	 = SA_RESTART;
  act.sa_handler = (void *)responseTimeout;
  sigaction(SIGVTALRM,  &act, NULL);

  /* cgi/ssi */
  act.sa_flags	 = SA_NOCLDSTOP;
  act.sa_handler = (void *)reapChild;
  sigaction(SIGCHLD, &act, NULL);

  /* ignore */
#if	DEBUG
  act.sa_handler = (void *)debugSignal;
  sigaction(SIGPIPE, &act, NULL);
#else
  act.sa_handler = SIG_IGN;
  sigaction(SIGPIPE, &act, NULL);
#endif	/* DEBUG */
}

/*
 * Function Name:	trap
 * Description:		trap all signals
 * Arguments:		signal
 * Return Value:	never return
 */
static void
trap(signo)
	int	signo;
{
  httpd->trap.signo = signo;
  switch (signo) {
    case SIGINT:
    case SIGQUIT:
    case SIGTERM:
      httpd->trap.sigexit = TRUE;
      break;
    case SIGBUS:
    case SIGSEGV:
      wvSyslog(LOG_CRIT, "killed by sig(%d).. it's a BUG!\n", signo);
      quit();
    case SIGHUP:
      httpd->trap.sighup = TRUE;
      break;
    default:
      break;
  }
}

/*****************************************************************************/

/*
 * Function Name:	reapChild
 * Description:		wait for child process termination
 * Arguments:		none
 * Return Value:	none
 */
static void
reapChild(void)
{
  pid_t			pid;
  int			status;
  register Thread *	actives = threadActives();

  while ((pid = waitpid((pid_t)0, &status, WNOHANG)) > 0) {
    register Thread *	tp;

    for (tp = actives; tp != NULL; tp = tp->next) {
      register Message *	mp;

      for (mp = &tp->mqueue; mp != NULL; mp = mp->next) {
	if (mp->cgi.pid == pid) {
	  if (mp->status == MSG_WAITING) {
	    mp->status = MSG_PROCESSING;
#if	DEBUG
	    fprintf(stderr, "[signal] next status PROCESSING\n");
#endif	/* DEBUG */
	  }
	  else {
	    mp->status = MSG_LOGGING;
#if	DEBUG
	    fprintf(stderr, "[signal] next status LOGGING\n");
#endif	/* DEBUG */
	  }
#if	DEBUG
      	  fprintf(stderr, "[signal] pid=%d, status=%d\n",
		  pid, WEXITSTATUS(status));
#endif	/* DEBUG */
    	}
      }
    }
  }
}

#if	DEBUG
/*
 * Function Name:	debugSignal
 * Description:		debug function for signal handler
 * Arguments:		none
 * Return Value:	none
 */
static void
debugSignal(void)
{
  fputs("[signal]: caught a signal (PIPE)\n", stderr);
}
#endif	/* DEBUG */

