/*****************************************************************************
 *
 * FILE:	ssl.c
 * DESCRIPTION:	secure socket layer module for Wyvern
 * DATE:	Thu, Jan 20 2000
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.com/cover.html
 * COPYRIGHT:	(C) 2000-2003 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *   Copyright (C) 2000-2003 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *   All rights reserved.
 *  
 *   Redistribution and use in source and binary forms, with or without
 *   modification, are permitted provided that the following conditions
 *   are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in
 *      the documentation and/or other materials provided with the
 *      distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY KOUICHI ABE (WALL) ``AS IS'' AND ANY
 *   EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL KOUICHI ABE (WALL) BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *   BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 *   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 *   OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 *   IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: ssl.c,v 1.26.2.15 2003/11/28 16:06:51 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#if	HAVE_FCNTL_H
#include <fcntl.h>
#endif	/* HAVE_FCNTL_H */
#if	HAVE_SYS_FILIO_H
#include <sys/filio.h>
#else
#if	HAVE_SYS_IOCTL_H
#include <sys/ioctl.h>
#endif	/* HAVE_SYS_IOCTL_H */
#endif	/* HAVE_SYS_FILIO_H */
#include <sys/types.h>
#if	HAVE_SYS_SOCKET_H
#include <sys/socket.h>
#endif	/* HAVE_SYS_SOCKET_H */
#if	HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif	/* HAVE_NETINET_IN_H */
#if	HAVE_ARPA_INET_H
#include <arpa/inet.h>
#endif	/* HAVE_ARPA_INET_H */
#if	HAVE_NETDB_H
#include <netdb.h>
#endif	/* HAVE_NETDB_H */
#include <errno.h>
#include <openssl/ssl.h>
#include <openssl/err.h>
#include "wyvern.h"
#include "functions.h"


/*****************************************************************************
 *
 *	definition of macros and structures
 *
 *****************************************************************************/
#define	SSL_DIR			SERVER_ROOT "/ssl"
#define	SSL_SERVER_CRT		SSL_DIR "/wyvern.crt"
#define	SSL_SERVER_KEY		SSL_DIR "/wyvern.key"
#define	DH_PARAM_512_PEM	SSL_DIR "/dh_param_512.pem"
#define	DH_PARAM_1024_PEM	SSL_DIR "/dh_param_1024.pem"
#define	SSL_CIPHER_SUITE \
	"ALL:!ADH:!EXPORT56:RC4+RSA:+HIGH:+MEDIUM:+LOW:+SSLv2:+EXP:+eNULL"


/*****************************************************************************
 *
 *	declaration of local functions
 *
 *****************************************************************************/
static int	init_rsa_keys(void);
static RSA *	tmp_rsa_callback(SSL *, int, int);
static int	init_dh_keys(void);
static DH *	tmp_dh_callback(SSL *, int, int);


/*****************************************************************************
 *
 *	declaration of local variables
 *
 *****************************************************************************/
#define	WYVERN_SESSION_ID	"56e512050ff6639de2123e40424281fc"
static unsigned char *	sid_ctx = (unsigned char *)WYVERN_SESSION_ID;
static RSA *		rsa_512	 = NULL;	/* set up ephemeral RSA stuff */
static RSA *		rsa_1024 = NULL;	/* set up ephemeral RSA stuff */
static DH *		dh_512	 = NULL;	/* set up ephemeral DH stuff */
static DH *		dh_1024	 = NULL;	/* set up ephemeral DH stuff */


/*****************************************************************************
 *
 *	definition of functions
 *
 *****************************************************************************/

/*
 * FUNCTION NAME:	sslOpenChannel
 * DESCRIPTION:		open SSL channel
 * ARGUMENTS:		none
 * RETURN VALUE:	SSL context
 */
SSL_CTX *
sslOpenChannel(void)
{
  SSL_CTX *	ssl_ctx;
  SSL_METHOD *	method = NULL;

  /* initialize SSL */
  SSL_library_init();		/* same as OpenSSL_add_ssl_algorithms(); */
  SSL_load_error_strings();

  /* create new SSL context */
#if	!defined(NO_SSL2) && !defined(NO_SSL3)
  method = SSLv23_server_method();
#elif	!defined(NO_SSL3)
  method = SSLv3_server_method();
#elif	!defined(NO_SSL2)
  method = SSLv2_server_method();
#endif
  ssl_ctx = (SSL_CTX *)SSL_CTX_new(method);
  if (ssl_ctx == NULL) {
    die("failed to create a SSL context\n");
  }

  /* set SSL Context options */
  SSL_CTX_set_options(ssl_ctx, SSL_OP_ALL);

  /* ceritification */
  if (SSL_CTX_use_certificate_chain_file(ssl_ctx, SSL_SERVER_CRT) != 1) {
    die("failed to load Certificate file: %s\n", SSL_SERVER_CRT);
  }
  if (SSL_CTX_use_PrivateKey_file(ssl_ctx, SSL_SERVER_KEY,
				  SSL_FILETYPE_PEM) != 1) {
    die("failed to load PrivateKey file: %s\n", SSL_SERVER_KEY);
  }
  if (SSL_CTX_check_private_key(ssl_ctx) != 1) {
    die("Private key does not match the certificate\n");
  }

  /* RSA keys for ephemeral key exchange */
  if (init_rsa_keys() == 0) {
    SSL_CTX_set_tmp_rsa_callback(ssl_ctx, tmp_rsa_callback);
  }
  /* DH keys for ephemeral key exchange */
  if (init_dh_keys() == 0) {
    SSL_CTX_set_tmp_dh_callback(ssl_ctx, tmp_dh_callback);
  }

#if	SSLEAY_VERSION_NUMBER >= 0x00906000L
  SSL_CTX_set_mode(ssl_ctx, SSL_MODE_ENABLE_PARTIAL_WRITE |
		   SSL_MODE_ACCEPT_MOVING_WRITE_BUFFER |
		   SSL_MODE_AUTO_RETRY);
  SSL_CTX_set_verify(ssl_ctx, SSL_VERIFY_NONE, NULL);
  SSL_CTX_set_session_cache_mode(ssl_ctx, SSL_SESS_CACHE_SERVER);
#endif

  /* set cipher list */
  if (SSL_CTX_set_cipher_list(ssl_ctx, SSL_CIPHER_SUITE) == 0) {
    die("failed to set cipher list\n");
  }

  SSL_CTX_set_session_id_context(ssl_ctx, sid_ctx, strlen(sid_ctx));
#if	DEBUG
  fprintf(stderr, "[ssl]\t%s\n\t%s\n", SSL_SERVER_CRT, SSL_SERVER_KEY);
  fprintf(stderr, "[ssl]\t%s\n", X509_get_default_cert_dir());
#endif	/* DEBUG */

  return ssl_ctx;
}

/*
 * FUNCTION NAME:	sslAccept
 * DESCRIPTION:		establish a SSL connection
 * ARGUMENTS:		thread
 * RETURN VALUE:	Successful: SSL_OK, error: SSL_ERROR, retry: SSL_RETRY
 */
int
sslAccept(thd)
	register Thread *	thd;
{
  int	ret;
  
  ret = SSL_accept(thd->conn.sock.ssl);
  switch (SSL_get_error(thd->conn.sock.ssl, ret)) {
    case SSL_ERROR_NONE:
      break;

    case SSL_ERROR_WANT_READ:
#if	DEBUG
      fprintf(stderr, "[SSL_accept(%s)] more read data = %d\n",
		thd->conn.client.port, SSL_want_read(thd->conn.sock.ssl));
#endif	/* DEBUG */
      return SSL_RETRY;

    case SSL_ERROR_SYSCALL:
      wvSyslog(LOG_ERR, "[ssl] some I/O error occurred\n");
      return SSL_ERROR;
    case SSL_ERROR_WANT_WRITE:
    case SSL_ERROR_WANT_X509_LOOKUP:
    case SSL_ERROR_ZERO_RETURN:
    case SSL_ERROR_SSL:
    default: 
#if	DEBUG
      fprintf(stderr, "[SSL_accept(%s)] SSL ERROR = %d: ret = %d\n",
		thd->conn.client.port,
		SSL_get_error(thd->conn.sock.ssl, ret), ret);
#endif	/* DEBUG */
      wvSyslog(LOG_ERR, "failed to accept for SSL connection\n");
      return SSL_ERROR;
  }
#if	DEBUG
  fprintf(stderr, "<sslNegotiation>...OK\tCC_CIPHER=%s:Version=%s;%s\n",
	  SSL_get_cipher(thd->conn.sock.ssl),
	  SSL_get_cipher_version(thd->conn.sock.ssl),
	  SSL_get_version(thd->conn.sock.ssl));
#endif	/* DEBUG */

  return SSL_OK;
}

/*****************************************************************************/

/*
 * FUNCTION NAME:	init_rsa_keys
 * DESCRIPTION:		initialize RSA keys
 * ARGUMENTS:		none
 * RETURN VALUE:	successful: 0, otherwise: -1
 */
static int
init_rsa_keys(void)
{
  /* initialize temporary rsa keys */
#if	SSLEAY_VERSION_NUMBER >= 0x0900
  rsa_512 = RSA_generate_key(512, RSA_F4, NULL, NULL);
#else
  rsa_512 = RSA_generate_key(512, RSA_F4, NULL);
#endif
  if (rsa_512 == NULL) {
    return -1;
  }

#if	SSLEAY_VERSION_NUMBER >= 0x0900
  rsa_1024 = RSA_generate_key(1024, RSA_F4, NULL, NULL);
#else
  rsa_1024 = RSA_generate_key(1024, RSA_F4, NULL);
#endif
  if (rsa_1024 == NULL) {
    return -1;
  }

  return 0;
}

/*
 * FUNCTION NAME:	tmp_rsa_callback
 * DESCRIPTION:		make a temporary private key
 * ARGUMENTS:		SSL connect, export flag, key length
 * RETURN VALUE:	key 512 / key 1024
 */
static RSA *
tmp_rsa_callback(s, is_export, keylength)
	SSL *	s;
	int	is_export;
	int	keylength;
{
  RSA *	rsa_tmp = NULL;

  switch (keylength) {
    case 512:
      rsa_tmp = rsa_512;
      break;
    case 1024:
      rsa_tmp = rsa_1024;
      break;
    default:
      /* generateing a key on the fly is very costly, so use what is there */
      if (rsa_1024) {
	rsa_tmp = rsa_1024;
      }
      else {
	rsa_tmp = rsa_512;	/* use at least a shorter key */
      }
      break;
  }

  return rsa_tmp;
}

/*
 * FUNCTION NAME:	init_dh_keys
 * DESCRIPTION:		initialize DH keys
 * ARGUMENTS:		none
 * RETURN VALUE:	successful: 0, otherwise: -1
 */
static int
init_dh_keys(void)
{
  register FILE *	fin;

  /* initialize temporary rsa keys */
  fin = wvSafeFOpen(DH_PARAM_512_PEM, "r");
  if (fin == NULL) {
    return -1;
  }
  dh_512 = PEM_read_DHparams(fin, NULL, NULL, NULL);
  fclose(fin);

  fin = wvSafeFOpen(DH_PARAM_1024_PEM, "r");
  if (fin == NULL) {
    return -1;
  }
  dh_1024 = PEM_read_DHparams(fin, NULL, NULL, NULL);
  fclose(fin);

  return 0;
}

/*
 * FUNCTION NAME:	tmp_dh_callback
 * DESCRIPTION:		make a temporary private key
 * ARGUMENTS:		SSL connect, export flag, key length
 * RETURN VALUE:	key 512 / key 1024
 */
static DH *
tmp_dh_callback(s, is_export, keylength)
	SSL *	s;
	int	is_export;
	int	keylength;
{
  DH *	dh_tmp = NULL;

  switch (keylength) {
    case 512:
      dh_tmp = dh_512;
      break;
    case 1024:
      dh_tmp = dh_1024;
      break;
    default:
      /* generateing a key on the fly is very costly, so use what is there */
      if (dh_1024) {
	dh_tmp = dh_1024;
      }
      else {
	dh_tmp = dh_512;	/* use at least a shorter key */
      }
      break;
  }

  return dh_tmp;
}

