/*****************************************************************************
 *
 * FILE:	logview.c
 * DESCRIPTION:	output an access.log with decoded escaped uri
 * DATE:	Thu, Apr 20 2000
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/cover.html
 * COPYRIGHT:	(C) 2000-2002 졿Kouichi ABE (WALL), All rights reserved.
 * HOWTOMAKE:	gcc -O2 -I/usr/local/include -L/usr/local/lib logview.c -lwkf
 * LICENSE:
 *  Copyright (C) 2000-2002 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *  
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in
 *      the documentation and/or other materials provided with the
 *      distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY KOUICHI ABE (WALL) ``AS IS'' AND ANY
 *   EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL KOUICHI ABE (WALL) BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *   BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 *   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 *   OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 *   IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: logview.c,v 1.3.6.3 2003/05/09 05:57:01 kouichi Exp $
 *
 *****************************************************************************/

#include <stdio.h>
#include <string.h>
#define	HAVE_WKF_H	1
#if	HAVE_WKF_H
#include <wkf.h>
#endif	/* HAVE_WKF_H */

#define	EOL	'\0'

#ifndef	_STRING_T
#define	_STRING_T	1
typedef	char *	String;
#endif	/* _STRING_T */

static String	decodeEscapedURI(String);
static void	parser(FILE *);

/*
 * Function Name:	decodeEscapedURI
 * Description:		escaped character is decoded
 * Arguments:		string
 * Return Value:	decoded string
 */
static String
decodeEscapedURI(s)
	String	s;
{
  register String	os  = s;
  register String	ret = s;
  char			c, c1;

  while ((c = *os) != EOL) {
    if (c == '%') {
      os++;
      if ((c = *os++) && (c1 = *os++)) {
	*s++ =	((c  >= 'A' ? ((c  & 0xdf) - 'A') + 10 : c  - '0') * 16) +
		((c1 >= 'A' ? ((c1 & 0xdf) - 'A') + 10 : c1 - '0'));
      }
      else {
        return ret;
      }
    }
    else {
      *s++ = c;
      os++;
    }
  }
  *s = EOL;

  return ret;
}

/*
 * Function Name:	parser
 * Description:		frontend for decodeEscapedURI
 * Arguments:		file pointer
 * Return Value:	none
 */
static void
parser(fin)
	register FILE *	fin;
{
  char	buf[BUFSIZ];
#if	HAVE_WKF_H
  char	ret[BUFSIZ*2];
#endif	/* HAVE_WKF_H */

  memset(buf, 0, sizeof(buf));
  while (fgets(buf, sizeof(buf), fin) != NULL) {
#if	HAVE_WKF_H
    decodeEscapedURI(buf);
#if	WKF_RELEASE
    if (wkfConvertKanjiCodeOfString(KC_UNKNOWN, buf, KC_EUC, ret, sizeof(ret)) == CONV_OK) {
      fputs(ret, stdout);
    }
    else {
      fputs(buf, stdout);
    }
#else
    if (wkfConvertKanjiCodeOfString(KC_EUC, buf, ret) == CONV_OK) {
      fputs(ret, stdout);
    }
    else {
      fputs(buf, stdout);
    }
#endif	/* WKF_RELEASE */
    memset(buf, 0, sizeof(buf));
    memset(ret, 0, sizeof(ret));
#else
    fputs(decodeEscapedURI(buf), stdout);
#endif	/* HAVE_WKF_H */
  }
}

/*
 *	MAIN
 */
int
main(argc, argv)
	int	argc;
	char *	argv[];
{
  register FILE *	fin;

  if (argc == 1) {
    fin = stdin;
    parser(fin);
  }
  else {
    int	i;

    for (i = 1; i < argc; i++) {
      fin = fopen(*++argv, "r");
      if (fin == NULL) {
	continue;
      }
      parser(fin);
      fclose(fin);
    }
  }

  return 0;
}

