/*****************************************************************************
 *
 * FILE:	wpasswd.c
 * DESCRIPTION:	a front end program for '.wpasswd'
 * DATE:	Sat, Dec 12 1999
 * UPDATE:	Sun, Dec 19 2004
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/cover.html
 * COPYRIGHT:	(c) 1999-2004 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *  Copyright (c) 1999-2004 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *  
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in
 *      the documentation and/or other materials provided with the
 *      distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY KOUICHI ABE (WALL) ``AS IS'' AND ANY
 *   EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL KOUICHI ABE (WALL) BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *   BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 *   WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 *   OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 *   IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: wpasswd.c,v 1.8.6.4 2004/12/19 13:55:32 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_STDLIB_H
#include <stdlib.h>
#endif	/* HAVE_STDLIB_H */
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#if	HAVE_FCNTL_H
#include <fcntl.h>
#endif	/* HAVE_FCNTL_H */
#if	HAVE_DB_H
#include <db.h>
#endif	/* HAVE_DB_H */
#if	HAVE_NDBM_H
#include <ndbm.h>
#endif	/* HAVE_NDBM_H */
#if	HAVE_MD5_H
#include <md5.h>
#endif	/* HAVE_MD5_H */
#include <sys/stat.h>
#if	HAVE_TERMIOS_H
#include <termios.h>
#endif	/* HAVE_TERMIOS_H */
#include <errno.h>


/*****************************************************************************
 *
 *	Copyright
 *
 *****************************************************************************/
const char	Program[]   = "wpasswd";
const char	Copyright[] =
	"Copyright (c) 1999-2004 Kouichi ABE (WALL), All rights reserved.";


/*****************************************************************************
 *
 *	declaration of macros and structures
 *
 *****************************************************************************/
#define	EOL	'\0'
#define CR	'\015'
#define LF	'\012'

typedef char *	String;
typedef enum {
  DB_PLAIN,	/* plain file */
  DB_DB,	/* berkeley db */
  DB_DBM	/* unix ndbm */
} db_t;


/*****************************************************************************
 *
 *	declaration of local functions
 *
 *****************************************************************************/
#ifndef	HAVE_LCHMOD
static int	lchmod(const char *, mode_t);
#endif	/* HAVE_LCHMOD */
static String	toMD5(String);
static String	chomp(String);
static String	getPassword(void);
static int	passwordWithDB(const String, const String);
static int	passwordWithDBM(const String, const String);
static int	passwordWithPlain(const String, const String);
static void	usage(void);


/*****************************************************************************
 *
 *	definition of functions
 *
 *****************************************************************************/

#ifndef	HAVE_LCHMOD
static int
lchmod(path, mode)
	const char *	path;
	mode_t		mode;
{
  struct stat	sbuf;

  if (lstat(path, &sbuf) == 0 && S_ISLNK(sbuf.st_mode)) {
    errno = ENOSYS;
    return -1;
  }

  return chmod(path, mode);
}
#endif	/* HAVE_LCHMOD */

/*
 * FUNCTION NAME:	toMD5
 * DESCRIPTION:		make a new password
 * ARGUMENTS:		password (plain text)
 * RETURN VALUE:	a crypted password
 */
static String
toMD5(text)
	const String	text;
{
  MD5_CTX	md5_ctx;
  unsigned char	md5[MD5_DIGEST_LENGTH];
  unsigned char	md5_hex[MD5_DIGEST_LENGTH*2+1];
  int		i;

  MD5Init(&md5_ctx);
  MD5Update(&md5_ctx, text, strlen(text));
  MD5Final(md5, &md5_ctx);

  memset(md5_hex, 0, sizeof(md5_hex));
  for (i = 0; i < MD5_DIGEST_LENGTH; i++) {
    (void)snprintf(md5_hex + i * 2, sizeof(md5_hex) - 1, "%02x", md5[i]);
  }

  return strdup(md5_hex);
}

/*
 * Function Name:	chomp
 * Description:		remove  CRLF, CR or LF from string end
 * Arguments:		string
 * Return Value:	string
 */
static String
chomp(s)
	String	s;
{
  String	p;
  size_t	len;

  if (s == NULL) {
    return NULL;
  }

  len = strlen(s);
  p   = &s[len - 1];
  if (*p == LF) {
    *p-- = EOL;
  }
  if (*p == CR) {
    *p-- = EOL;
  }

  return s;
}


/*
 * FUNCTION NAME:	getPassword
 * DESCRIPTION:		recept user and password
 * ARGUMENTS:		none
 * RETURN VALUE:	password string with plain text
 */
static String
getPassword(void)
{
  struct termios	tc_old;
  struct termios	tc_new;

  tcgetattr(fileno(stdin), &tc_old);
  for (;;) {
    char	pwd1[BUFSIZ];
    char	pwd2[BUFSIZ];
    size_t	len;

    memset(pwd1, 0, sizeof(pwd1));
    fputs("Password: ", stdout);
    tc_new	    = tc_old;
    tc_new.c_lflag &= ~ECHO;
    tcsetattr(fileno(stdin), TCSAFLUSH, &tc_new);
    if (fgets(pwd1, sizeof(pwd1), stdin) == NULL) {
      exit(0);
    }
    putchar(LF);

    memset(pwd2, 0, sizeof(pwd2));
    fputs("Retype password: ", stdout);
    if (fgets(pwd2, sizeof(pwd2), stdin) == NULL) {
      exit(0);
    }
    putchar(LF);
    tcsetattr(fileno(stdin), TCSANOW, &tc_old);

    len = strlen(pwd1);
    if (len == 1) {
      fputs("Empty Password; try again\n", stderr);
    }
    else if (strncmp(pwd1, pwd2, len) == 0) {
      return strdup(chomp(pwd1));
    }
    else {
      fputs("Mismatch; try again\n", stderr);
    }
  }
}

/*
 * FUNCTION NAME:	passwordWithDB
 * DESCRIPTION:		register username and password
 * ARGUMENTS:		password file, username
 * RETURN VALUE:	successful: 0, otherwise: -1
 */
static int
passwordWithDB(password_file, username)
	const String	password_file;
	const String	username;
{
  String	password = toMD5(getPassword());
  register DB *	dp;
  DBT		key;
  DBT		val;
  int		ret;

  if (password_file == NULL || username == NULL || password == NULL) {
    return -1;
  }

  dp = dbopen(password_file, O_RDWR|O_CREAT, 0600, DB_HASH, NULL);
  if (dp == NULL) {
    fprintf(stderr, "ERROR: failed to open %s\n", password_file);
    return -1;
  }

  key.data = username;
  key.size = strlen(username);
  val.data = password;
  val.size = strlen(password);

  ret = dp->put(dp, &key, &val, 0);

  dp->close(dp);

  return ret;
}

/*
 * FUNCTION NAME:	passwordWithDBM
 * DESCRIPTION:		register username and password
 * ARGUMENTS:		password file, username
 * RETURN VALUE:	successful: 0, otherwise: -1
 */
static int
passwordWithDBM(password_file, username)
	const String	password_file;
	const String	username;
{
  String		password = toMD5(getPassword());
  register DBM *	dp;
  datum			key;
  datum			val;
  int			ret;

  if (password_file == NULL || username == NULL || password == NULL) {
    return -1;
  }

  dp = dbm_open(password_file, O_RDWR|O_CREAT, 0600);
  if (dp == NULL) {
    fprintf(stderr, "ERROR: failed to open %s\n", password_file);
    return -1;
  }

  key.dptr  = username;
  key.dsize = strlen(username);
  val.dptr  = password;
  val.dsize = strlen(password);

  ret = dbm_store(dp, key, val, DBM_REPLACE);

  dbm_close(dp);

  return ret;
}

/*
 * FUNCTION NAME:	passwordWithPlain
 * DESCRIPTION:		register username and password
 * ARGUMENTS:		password file, username
 * RETURN VALUE:	successful: 0, otherwise: -1
 */
static int
passwordWithPlain(password_file, username)
	const String	password_file;
	const String	username;
{
  String		password = toMD5(getPassword());
  register FILE *	dp;

  if (password_file == NULL || username == NULL || password == NULL) {
    return -1;
  }

  dp = fopen(password_file, "a");
  if (dp == NULL) {
    fprintf(stderr, "ERROR: failed to open %s\n", password_file);
    return -1;
  }
  fprintf(dp, "%s:%s\n", username, password);
  fclose(dp);
  lchmod(password_file, 0600);

  return 0;
}

/*
 * FUNCTION NAME:	usage
 * DESCRIPTION:		print usage
 * ARGUMENTS:		none
 * RETURN VALUE:	none
 */
static void
usage(void)
{
  fprintf(stderr,
	  "%s\n"
	  "usage: %s [-db|-dbm] passwordfile username\n",
	  Copyright, Program);
  exit(64);
}


/*
 *	MAIN FUNCTION
 */
int
main(argc, argv)
	int	argc;
	char *	argv[];
{
  db_t		db_type	= DB_PLAIN;
  String	password_file;
  String	username;

  /* parsing arguments */
  switch (argc) {
    case 4:
      if (strncmp(argv[1], "-db", 3) == 0) {
	db_type = DB_DB;
      }
      else if (strncmp(argv[1], "-dbm", 4) == 0) {
	db_type = DB_DBM;
      }
      else {
	usage();
      }
      password_file = argv[2];
      username	    = argv[3];
      break;
    case 3:
      password_file = argv[1];
      username	    = argv[2];
      break;
    default:
      usage();
  }

  switch (db_type) {
    case DB_DB:
      passwordWithDB(password_file, username);
      break;
    case DB_DBM:
      passwordWithDBM(password_file, username);
      break;
    default:
      passwordWithPlain(password_file, username);
      break;
  }

  return 0;
}

